@@ -78,6 +78,8 @@
      ,@forms
      (dolist (f wrapped-functions) (unencapsulate f '!cold-init))))
 
+(defun !c-runtime-noinform-p () (/= (extern-alien "lisp_startup_options" char) 0))
+
 ;;; called when a cold system starts up
 (defun !cold-init (&aux real-choose-symbol-out-fun)
   "Give the world a shove and hope it spins."
@@ -90,13 +92,19 @@
   (setq *error-output* (!make-cold-stderr-stream)
                       *standard-output* *error-output*
                       *trace-output* *error-output*)
-  (write-string "COLD-INIT... ")
+  (unless (!c-runtime-noinform-p)
+    (write-string "COLD-INIT... "))
 
   ;; Assert that FBOUNDP doesn't choke when its answer is NIL.
   ;; It was fine if T because in that case the legality of the arg is certain.
   ;; And be extra paranoid - ensure that it really gets called.
   (locally (declare (notinline fboundp)) (fboundp '(setf !zzzzzz)))
 
+  ;; Ensure that *CURRENT-THREAD* and *HANDLER-CLUSTERS* have sane values.
+  ;; create_thread_struct() assigned NIL/unbound-marker respectively.
+  (sb!thread::init-initial-thread)
+  (show-and-call sb!kernel::!target-error-cold-init)
+
   ;; Putting data in a synchronized hashtable (*PACKAGE-NAMES*)
   ;; requires that the main thread be properly initialized.
   (show-and-call thread-init-or-reinit)
@@ -167,14 +175,17 @@
   (!with-init-wrappers
    (dolist (x *!cold-defuns*)
      (destructuring-bind (name . inline-expansion) x
-       (%defun name (fdefinition name) nil inline-expansion))))
+       (%defun name (fdefinition name) inline-expansion))))
 
   ;; KLUDGE: Why are fixups mixed up with toplevel forms? Couldn't
   ;; fixups be done separately? Wouldn't that be clearer and better?
   ;; -- WHN 19991204
   (/show0 "doing cold toplevel forms and fixups")
-  (progn (write `("Length(TLFs)= " ,(length *!cold-toplevels*)))
-         (terpri))
+  (unless (!c-runtime-noinform-p)
+    (write `("Length(TLFs)=" ,(length *!cold-toplevels*)) :escape nil)
+    (terpri))
+  ;; only the basic external formats are present at this point.
+  (setq sb!impl::*default-external-format* :latin-1)
 
   (!with-init-wrappers
     (loop for index-in-cold-toplevels from 0
@@ -271,6 +282,8 @@
   (/show0 "enabling internal errors")
   (setf (extern-alien "internal_errors_enabled" int) 1)
 
+  (show-and-call sb!disassem::!compile-inst-printers)
+
   ;; Toggle some readonly bits
   (dovector (sc sb!c:*backend-sc-numbers*)
     (when sc
@@ -289,18 +302,15 @@
   (/show0 "back from first GC")
 
   ;; The show is on.
-  (terpri)
   (/show0 "going into toplevel loop")
   (handling-end-of-the-world
     (toplevel-init)
     (critically-unreachable "after TOPLEVEL-INIT")))
 
-(define-deprecated-function :early "1.0.56.55" quit (exit sb!thread:abort-thread)
-    (&key recklessly-p (unix-status 0))
-  (if (or recklessly-p (sb!thread:main-thread-p))
-      (exit :code unix-status :abort recklessly-p)
-      (sb!thread:abort-thread))
-  (critically-unreachable "after trying to die in QUIT"))
+(defun quit (&key recklessly-p (unix-status 0))
+  "Calls (SB-EXT:EXIT :CODE UNIX-STATUS :ABORT RECKLESSLY-P),
+see documentation for SB-EXT:EXIT."
+  (exit :code unix-status :abort recklessly-p))
 
 (declaim (ftype (sfunction (&key (:code (or null exit-code))
                                  (:timeout (or null real))
@@ -356,7 +366,6 @@ process to continue normally."
 ;;;; initialization functions
 
 (defun thread-init-or-reinit ()
-  (sb!thread::init-initial-thread)
   (sb!thread::init-job-control)
   (sb!thread::get-foreground))
 
@@ -367,6 +376,11 @@ process to continue normally."
   (setf sb!alien::*default-c-string-external-format* nil)
   ;; WITHOUT-GCING implies WITHOUT-INTERRUPTS.
   (without-gcing
+    (finalizers-reinit)
+    ;; Create *CURRENT-THREAD* first, since initializing a stream calls
+    ;; ALLOC-BUFFER which calls FINALIZE which acquires **FINALIZER-STORE-LOCK**
+    ;; which needs a valid thread in order to grab a mutex.
+    (sb!thread::init-initial-thread)
     ;; Initialize streams first, so that any errors can be printed later
     (stream-reinit t)
     (os-cold-init-or-reinit)
@@ -431,15 +445,11 @@ process to continue normally."
     (%cold-print x 0))
   (values))
 
-(in-package "SB!INT")
-(defun !unintern-symbols ()
-    ;; For some reason uninterning these:
-    ;;    DEF!TYPE DEF!CONSTANT DEF!STRUCT
-    ;; does not work, they stick around as uninterned symbols.
-    ;; Some other macros must expand into them. Ugh.
+(push
   '("SB-INT"
-    defenum defun-cached with-globaldb-name
+    defenum defun-cached with-globaldb-name def!type def!struct
     .
     #!+sb-show ()
     #!-sb-show (/hexstr /nohexstr /noshow /noshow0 /noxhow
-                /primitive-print /show /show0 /xhow)))
+                /primitive-print /show /show0 /xhow))
+  sb!impl::*!removable-symbols*)
