@@ -110,15 +110,27 @@ evaluated as a PROGN."
   `(if ,test nil ,(prognify forms env))))
 
 (sb!xc:defmacro and (&rest forms)
-  (named-let expand-forms ((nested nil) (forms forms))
+  (named-let expand-forms ((nested nil) (forms forms) (ignore-last nil))
     (cond ((endp forms) t)
           ((endp (rest forms))
-           ;; Preserve non-toplevelness of the form!
-           (let ((car (car forms))) (if nested car `(the t ,car))))
+           (let ((car (car forms)))
+             (cond (nested
+                    car)
+                   (t
+                    ;; Preserve non-toplevelness of the form!
+                    `(the t ,car)))))
+          ((and ignore-last
+                (endp (cddr forms)))
+           (car forms))
+          ;; Better code that way, since the result will only have two
+          ;; values, NIL or the last form, and the precedeing tests
+          ;; will only be used for jumps
+          ((and (not nested) (cddr forms))
+           `(if ,(expand-forms t forms t)
+                ,@(last forms)))
           (t
            `(if ,(first forms)
-                ,(expand-forms t (rest forms))
-                nil)))))
+                ,(expand-forms t (rest forms) ignore-last))))))
 
 (sb!xc:defmacro or (&rest forms)
   (named-let expand-forms ((nested nil) (forms forms))
@@ -205,8 +217,7 @@ evaluated as a PROGN."
       `(progn
          (eval-when (:compile-toplevel)
            (sb!c:%compiler-defun ',name ,inline-thing t))
-         (%defun ',name ,named-lambda (sb!c:source-location)
-                 ,@(and inline-thing (list inline-thing)))
+         (%defun ',name ,named-lambda ,@(and inline-thing (list inline-thing)))
          ;; This warning, if produced, comes after the DEFUN happens.
          ;; When compiling, there's no real difference, but when interpreting,
          ;; if there is a handler for style-warning that nonlocally exits,
@@ -217,23 +228,6 @@ evaluated as a PROGN."
                  (sb!c::warn-if-setf-macro ',name))
                ',name))))))
 
-#-sb-xc-host
-(defun %defun (name def source-location &optional inline-lambda)
-  (declare (type function def))
-  ;; should've been checked by DEFMACRO DEFUN
-  (aver (legal-fun-name-p name))
-  (sb!c:%compiler-defun name inline-lambda nil)
-  (when (fboundp name)
-    (warn 'redefinition-with-defun
-          :name name :new-function def :new-location source-location))
-  (setf (fdefinition name) def)
-  ;; %COMPILER-DEFUN doesn't do this except at compile-time, when it
-  ;; also checks package locks. By doing this here we let (SETF
-  ;; FDEFINITION) do the load-time package lock checking before
-  ;; we frob any existing inline expansions.
-  (sb!c::%set-inline-expansion name nil inline-lambda)
-  (sb!c::note-name-defined name :function)
-  name)
 
 ;;;; DEFVAR and DEFPARAMETER
 
@@ -269,30 +263,19 @@ evaluated as a PROGN."
                     ,@(and docp
                            `(',doc)))))
 
+(defun %compiler-defglobal (name always-boundp value assign-it-p)
+  (sb!xc:proclaim `(global ,name))
+  (when assign-it-p
+    (set-symbol-global-value name value))
+  (sb!c::process-variable-declaration
+   name 'always-bound
+   ;; don't "weaken" the proclamation if it's in fact always bound now
+   (if (eq (info :variable :always-bound name) :always-bound)
+       :always-bound
+       always-boundp)))
+
 (defun %compiler-defvar (var)
   (sb!xc:proclaim `(special ,var)))
-
-#-sb-xc-host
-(defun %defvar (var source-location &optional (val nil valp) (doc nil docp))
-  (%compiler-defvar var)
-  (when (and valp
-             (not (boundp var)))
-    (set var val))
-  (when docp
-    (setf (fdocumentation var 'variable) doc))
-  (when source-location
-    (setf (info :source-location :variable var) source-location))
-  var)
-
-#-sb-xc-host
-(defun %defparameter (var val source-location &optional (doc nil docp))
-  (%compiler-defvar var)
-  (set var val)
-  (when docp
-    (setf (fdocumentation var 'variable) doc))
-  (when source-location
-    (setf (info :source-location :variable var) source-location))
-  var)
 
 ;;;; iteration constructs
 
@@ -432,25 +415,6 @@ evaluated as a PROGN."
 
 ;;;; conditions, handlers, restarts
 
-;;; KLUDGE: we PROCLAIM these special here so that we can use restart
-;;; macros in the compiler before the DEFVARs are compiled.
-;;;
-;;; For an explanation of these data structures, see DEFVARs in
-;;; target-error.lisp.
-(sb!xc:proclaim '(special *handler-clusters* *restart-clusters*))
-
-;;; Generated code need not check for unbound-marker in *HANDLER-CLUSTERS*
-;;; (resp *RESTART-). To elicit this we must poke at the info db.
-;;; SB!XC:PROCLAIM SPECIAL doesn't advise the host Lisp that *HANDLER-CLUSTERS*
-;;; is special and so it rightfully complains about a SETQ of the variable.
-;;; But I must SETQ if proclaming ALWAYS-BOUND because the xc asks the host
-;;; whether it's currently bound.
-;;; But the DEFVARs are in target-error. So it's one hack or another.
-(setf (info :variable :always-bound '*handler-clusters*)
-      #+sb-xc :always-bound #-sb-xc :eventually)
-(setf (info :variable :always-bound '*restart-clusters*)
-      #+sb-xc :always-bound #-sb-xc :eventually)
-
 (sb!xc:defmacro with-condition-restarts
     (condition-form restarts-form &body body)
   "Evaluates the BODY in a dynamic environment where the restarts in the list
@@ -592,15 +556,22 @@ evaluated as a PROGN."
              (make-apply-and-return (clause-data)
                (destructuring-bind (name tag keywords lambda-list body) clause-data
                  (declare (ignore name keywords))
-                 `(,tag (return-from ,block-tag
-                          ,(cond ((null lambda-list)
-                                  `(progn ,@body))
-                                 ((and (null (cdr lambda-list))
-                                       (not (member (car lambda-list)
-                                                    '(&optional &key &aux))))
-                                  `(funcall (lambda ,lambda-list ,@body) ,temp-var))
-                                 (t
-                                  `(apply (lambda ,lambda-list ,@body) ,temp-var))))))))
+                 (multiple-value-bind (body declarations) (parse-body body nil)
+                   `(,tag (return-from ,block-tag
+                            ,(cond ((null lambda-list)
+                                    `(locally ,@declarations ,@body))
+                                   ((and (null (cdr lambda-list))
+                                         (not (member (car lambda-list)
+                                                      '(&optional &key &aux))))
+                                    `(funcall (lambda ,lambda-list
+                                                ,@declarations
+                                                (progn ,@body))
+                                              ,temp-var))
+                                   (t
+                                    `(apply (lambda ,lambda-list
+                                              ,@declarations
+                                              (progn ,@body))
+                                            ,temp-var)))))))))
       (let ((clauses-data (mapcar #'parse-clause clauses)))
         `(block ,block-tag
            (let ((,temp-var nil))
@@ -665,17 +636,15 @@ evaluated as a PROGN."
                       `(cons ,(case (car test)
                                 ((named-lambda function) test)
                                 (t `(load-time-value ,test t)))
-                             ,(if (typep handler '(cons (eql function)))
-                                  `(callable-cast 1 handler-bind ,handler)
+                             ,(locally
                                   ;; Regardless of lexical policy, never allow
                                   ;; a non-callable into handler-clusters.
-                                  `(let ((x ,handler))
-                                     (declare (optimize (safety 3)))
-                                     (callable-cast 1 handler-bind x)))))
+                                  (declare (optimize (safety 3)))
+                                `(the (function-designator (condition)) ,handler))))
                      ((info :function :info name) ; known
                       ;; This takes care of CONTINUE,ABORT,MUFFLE-WARNING.
                       ;; #' will be evaluated in the null environment.
-                      `(load-time-value (cons ,test (callable-cast 1 handler-bind #',name)) t))
+                      `(load-time-value (cons ,test (the (function-designator (condition)) #',name)) t))
                      (t
                       ;; For each handler specified as #'F we must verify
                       ;; that F is fboundp upon entering the binding scope.
@@ -690,7 +659,7 @@ evaluated as a PROGN."
                                         (find-or-create-fdefn ',name) t))))
                       ;; Resolve to an fdefn at load-time.
                       `(load-time-value
-                        (cons ,test (find-or-create-fdefn (callable-cast 1 handler-bind ',name)))
+                        (cons ,test (find-or-create-fdefn (the (function-designator (condition)) ',name)))
                         t)))))
 
            (const-list (items)
@@ -807,22 +776,23 @@ specification."
                      ,@(remove no-error-clause cases)))))))
         (let* ((local-funs nil)
                (annotated-cases
-                (mapcar (lambda (case)
-                          (with-current-source-form (case)
-                            (with-unique-names (tag fun)
-                              (destructuring-bind (type ll &body body) case
-                                (unless (and (listp ll)
-                                             (symbolp (car ll))
-                                             (null (cdr ll)))
-                                  (error "Malformed HANDLER-CASE lambda-list. Should be either () or (symbol), not ~s."
-                                         ll))
-
-                                (push `(,fun ,ll ,@body) local-funs)
-                                (list tag type ll fun)))))
-                        cases)))
+                 (mapcar (lambda (case)
+                           (with-current-source-form (case)
+                             (with-unique-names (tag fun)
+                               (destructuring-bind (type ll &body body) case
+                                 (unless (and (listp ll)
+                                              (symbolp (car ll))
+                                              (null (cdr ll)))
+                                   (error "Malformed HANDLER-CASE lambda-list. Should be either () or (symbol), not ~s."
+                                          ll))
+                                 (multiple-value-bind (body declarations)
+                                     (parse-body body nil)
+                                   (push `(,fun ,ll ,@declarations (progn ,@body)) local-funs))
+                                 (list tag type ll fun)))))
+                         cases)))
           (with-unique-names (block cell form-fun)
             `(dx-flet ((,form-fun ()
-                         #!-x86 ,form
+                         #!-x86 (progn ,form) ;; no declarations are accepted
                          ;; Need to catch FP errors here!
                          #!+x86 (multiple-value-prog1 ,form (float-wait)))
                        ,@(reverse local-funs))
