@@ -20,7 +20,7 @@
   "The absolute pathname of the running SBCL runtime.")
 
 ;;; something not EQ to anything we might legitimately READ
-(defglobal *eof-object* (make-symbol "EOF-OBJECT"))
+(define-load-time-global *eof-object* (make-symbol "EOF-OBJECT"))
 
 (eval-when (:compile-toplevel :load-toplevel :execute)
   (defconstant max-hash sb!xc:most-positive-fixnum))
@@ -28,6 +28,11 @@
 (def!type hash ()
   `(integer 0 ,max-hash))
 
+;;; A number that can represent an index into a vector, including
+;;; one-past-the-end
+(deftype array-range ()
+  '(integer 0 #.sb!xc:array-dimension-limit))
+
 ;;; a type used for indexing into sequences, and for related
 ;;; quantities like lengths of lists and other sequences.
 ;;;
@@ -76,7 +81,7 @@
 ;;; The smallest power of two that is equal to or greater than X.
 (declaim (inline power-of-two-ceiling))
 (defun power-of-two-ceiling (x)
-  (declare (index x))
+  (declare (type index x))
   (ash 1 (integer-length (1- x))))
 
 (def!type load/store-index (scale lowtag min-offset
@@ -113,24 +118,6 @@
                                                           data-offset)
         `(integer ,min ,max)))))
 
-;;; the default value used for initializing character data. The ANSI
-;;; spec says this is arbitrary, so we use the value that falls
-;;; through when we just let the low-level consing code initialize
-;;; all newly-allocated memory to zero.
-;;;
-;;; KLUDGE: It might be nice to use something which is a
-;;; STANDARD-CHAR, both to reduce user surprise a little and, probably
-;;; more significantly, to help SBCL's cross-compiler (which knows how
-;;; to dump STANDARD-CHARs). Unfortunately, the old CMU CL code is
-;;; shot through with implicit assumptions that it's #\NULL, and code
-;;; in several places (notably both DEFUN MAKE-ARRAY and DEFTRANSFORM
-;;; MAKE-ARRAY) would have to be rewritten. -- WHN 2001-10-04
-(eval-when (:compile-toplevel :load-toplevel :execute)
-  ;; an expression we can use to construct a DEFAULT-INIT-CHAR value
-  ;; at load time (so that we don't need to teach the cross-compiler
-  ;; how to represent and dump non-STANDARD-CHARs like #\NULL)
-  (defparameter *default-init-char-form* '(code-char 0)))
-
 ;;; CHAR-CODE values for ASCII characters which we care about but
 ;;; which aren't defined in section "2.1.3 Standard Characters" of the
 ;;; ANSI specification for Lisp
@@ -769,7 +756,7 @@ NOTE: This interface is experimental and subject to change."
                  (values ,@result-temps))))))
     `(progn
        (pushnew ',var-name *cache-vector-symbols*)
-       (defglobal ,var-name nil)
+       (define-load-time-global ,var-name nil)
        ,@(when *profile-hash-cache*
            `((declaim (type (simple-array fixnum (3)) ,statistics-name))
              (defvar ,statistics-name)))
@@ -1034,27 +1021,6 @@ NOTE: This interface is experimental and subject to change."
   (def-constantly-fun constantly-nil nil)
   (def-constantly-fun constantly-0 0))
 
-;;; If X is a symbol, see whether it is present in *FEATURES*. Also
-;;; handle arbitrary combinations of atoms using NOT, AND, OR.
-(defun featurep (x)
-  (typecase x
-    (cons
-     (case (car x)
-       ((:not not)
-        (cond
-          ((cddr x)
-           (error "too many subexpressions in feature expression: ~S" x))
-          ((null (cdr x))
-           (error "too few subexpressions in feature expression: ~S" x))
-          (t (not (featurep (cadr x))))))
-       ((:and and) (every #'featurep (cdr x)))
-       ((:or or) (some #'featurep (cdr x)))
-       (t
-        (error "unknown operator in feature expression: ~S." x))))
-    (symbol (not (null (memq x *features*))))
-    (t
-      (error "invalid feature expression: ~S" x))))
-
 
 ;;;; utilities for two-VALUES predicates
 
@@ -1233,14 +1199,6 @@ NOTE: This interface is experimental and subject to change."
     (+ digits (floor (1- digits) comma-interval))))
 
 
-;;;; etc.
-
-;;; Given a pathname, return a corresponding physical pathname.
-(defun physicalize-pathname (possibly-logical-pathname)
-  (if (typep possibly-logical-pathname 'logical-pathname)
-      (translate-logical-pathname possibly-logical-pathname)
-      possibly-logical-pathname))
-
 ;;;; Deprecating stuff
 
 (deftype deprecation-state ()
@@ -1447,10 +1405,11 @@ NOTE: This interface is experimental and subject to change."
   #+sb-xc-host (declare (ignore software version name replacement-spec))
   #-sb-xc-host
   (setf (fdefinition name)
-        (sb!impl::set-closure-name
+        (set-closure-name
          (lambda (&rest args)
            (declare (ignore args))
            (deprecation-error software version 'function name replacement-spec))
+         t
          name)))
 
 (defun setup-variable-in-final-deprecation
@@ -1607,13 +1566,19 @@ NOTE: This interface is experimental and subject to change."
 (defmacro with-sane-io-syntax (&body forms)
   `(call-with-sane-io-syntax (lambda () ,@forms)))
 
+(!defvar *print-vector-length* nil
+  "Like *PRINT-LENGTH* but works on strings and bit-vectors.
+Does not affect the cases that are already controlled by *PRINT-LENGTH*")
+(declaim (always-bound *print-vector-length*))
+
 (defun call-with-sane-io-syntax (function)
   (declare (type function function))
   (macrolet ((true (sym)
                `(and (boundp ',sym) ,sym)))
     (let ((*print-readably* nil)
           (*print-level* (or (true *print-level*) 6))
-          (*print-length* (or (true *print-length*) 12)))
+          (*print-length* (or (true *print-length*) 12))
+          (*print-vector-length* (or (true *print-vector-length*) 200)))
       (funcall function))))
 
 ;;; Returns a list of members of LIST. Useful for dealing with circular lists.
@@ -1638,7 +1603,7 @@ NOTE: This interface is experimental and subject to change."
   "Toggle between different evaluator implementations. If set to :COMPILE,
 an implementation of EVAL that calls the compiler will be used. If set
 to :INTERPRET, an interpreter will be used.")
-
+(declaim (always-bound *evaluator-mode*))
 ;; This is not my preferred name for this function, but chosen for harmony
 ;; with everything else that refers to these as 'hash-caches'.
 ;; Hashing is just one particular way of memoizing, and it would have been
@@ -1717,6 +1682,19 @@ to :INTERPRET, an interpreter will be used.")
              (sorted (stable-sort wrapped comparator :key #'cdr)))
         (map-into sorted #'car sorted))))
 
+(declaim (inline schwartzian-stable-sort-vector))
+(defun schwartzian-stable-sort-vector (vector comparator &key key)
+  (if (null key)
+      (stable-sort (copy-seq vector) comparator)
+      (let* ((key (if (functionp key)
+                      key
+                      (symbol-function key)))
+             (wrapped (map 'vector (lambda (x)
+                                     (cons x (funcall key x)))
+                           vector))
+             (sorted (stable-sort wrapped comparator :key #'cdr)))
+        (map-into sorted #'car sorted))))
+
 ;;; Just like WITH-OUTPUT-TO-STRING but doesn't close the stream,
 ;;; producing more compact code.
 (defmacro with-simple-output-to-string
