@@ -50,6 +50,10 @@
 
 (define-load-time-global *package-graph-lock* nil)
 
+(defmacro with-package-graph ((&key) &body forms)
+  `(flet ((thunk () ,@forms))
+     (declare (dynamic-extent #'thunk))
+     (call-with-package-graph #'thunk)))
 (defun call-with-package-graph (function)
   (declare (function function))
   ;; FIXME: Since name conflicts can be signalled while holding the
@@ -59,13 +63,86 @@
 
 ;;; a map from package names to packages
 (define-load-time-global *package-names* nil)
-(declaim (type hash-table *package-names*))
+(declaim (type info-hashtable *package-names*))
 
-(defmacro with-package-names ((names &key) &body body)
-  `(let ((,names *package-names*))
-     (with-locked-system-table (,names)
+(defmacro with-package-names ((table-var &key) &body body)
+  `(let ((,table-var *package-names*))
+     (sb!thread::with-recursive-system-lock ((info-env-mutex ,table-var))
        ,@body)))
 
+(defmethod make-load-form ((p package) &optional environment)
+  (declare (ignore environment))
+  `(find-undeleted-package-or-lose ,(package-name p)))
+
+;;;; iteration macros
+
+(defmacro with-package-iterator ((mname package-list &rest symbol-types) &body body)
+  "Within the lexical scope of the body forms, MNAME is defined via macrolet
+such that successive invocations of (MNAME) will return the symbols, one by
+one, from the packages in PACKAGE-LIST. SYMBOL-TYPES may be any
+of :INHERITED :EXTERNAL :INTERNAL."
+  ;; SYMBOL-TYPES should really be named ACCESSIBILITY-TYPES.
+  (when (null symbol-types)
+    (%program-error "At least one of :INTERNAL, :EXTERNAL, or :INHERITED must be supplied."))
+  (dolist (symbol symbol-types)
+    (unless (member symbol '(:internal :external :inherited))
+      (%program-error "~S is not one of :INTERNAL, :EXTERNAL, or :INHERITED."
+                      symbol)))
+  (with-unique-names (bits index sym-vec pkglist symbol kind)
+    (let ((state (list bits index sym-vec pkglist))
+          (select (logior (if (member :internal  symbol-types) 1 0)
+                          (if (member :external  symbol-types) 2 0)
+                          (if (member :inherited symbol-types) 4 0))))
+      `(multiple-value-bind ,state (package-iter-init ,select ,package-list)
+         (let (,symbol ,kind)
+           (macrolet
+               ((,mname ()
+                   '(if (eql 0 (multiple-value-setq (,@state ,symbol ,kind)
+                                 (package-iter-step ,@state)))
+                        nil
+                        (values t ,symbol ,kind
+                                (car (truly-the list ,pkglist))))))
+             ,@body))))))
+
+(eval-when (:compile-toplevel :load-toplevel :execute)
+  (defun expand-pkg-iterator (range var body result-form)
+    (multiple-value-bind (forms decls) (parse-body body nil)
+      (with-unique-names (iterator winp next)
+        `(block nil
+           (with-package-iterator (,iterator ,@range)
+             (tagbody
+                  ,next
+                  (multiple-value-bind (,winp ,var) (,iterator)
+                    (declare (ignorable ,var))
+                    ,@decls
+                    (if ,winp
+                        (tagbody ,@forms (go ,next))
+                        (return ,result-form))))))))))
+
+(defmacro do-symbols ((var &optional (package '*package*) result-form)
+                           &body body-decls)
+  "DO-SYMBOLS (VAR [PACKAGE [RESULT-FORM]]) {DECLARATION}* {TAG | FORM}*
+   Executes the FORMs at least once for each symbol accessible in the given
+   PACKAGE with VAR bound to the current symbol."
+  (expand-pkg-iterator `((find-undeleted-package-or-lose ,package)
+                       :internal :external :inherited)
+                       var body-decls result-form))
+
+(defmacro do-external-symbols ((var &optional (package '*package*) result-form)
+                                    &body body-decls)
+  "DO-EXTERNAL-SYMBOLS (VAR [PACKAGE [RESULT-FORM]]) {DECL}* {TAG | FORM}*
+   Executes the FORMs once for each external symbol in the given PACKAGE with
+   VAR bound to the current symbol."
+  (expand-pkg-iterator `((find-undeleted-package-or-lose ,package) :external)
+                       var body-decls result-form))
+
+(defmacro do-all-symbols ((var &optional result-form) &body body-decls)
+  "DO-ALL-SYMBOLS (VAR [RESULT-FORM]) {DECLARATION}* {TAG | FORM}*
+   Executes the FORMs once for each symbol in every package with VAR bound
+   to the current symbol."
+  (expand-pkg-iterator '((list-all-packages) :internal :external)
+                       var body-decls result-form))
+
 ;;;; PACKAGE-HASHTABLE stuff
 
 (defmethod print-object ((table package-hashtable) stream)
@@ -125,10 +202,8 @@
           (package-hashtable-free table) (package-hashtable-free temp-table)
           (package-hashtable-deleted table) 0)))
 
-;;;; package locking operations, built conditionally on :sb-package-locks
+;;;; package locking operations, built unconditionally now
 
-#!+sb-package-locks
-(progn
 (defun package-locked-p (package)
   "Returns T when PACKAGE is locked, NIL otherwise. Signals an error
 if PACKAGE doesn't designate a valid package."
@@ -259,8 +334,6 @@ error if any of PACKAGES is not a valid package designator."
                 list)
               sb!c::*disabled-package-locks*)))
 
-) ; progn
-
 ;;;; more package-locking these are NOPs unless :sb-package-locks is
 ;;;; in target features. Cross-compiler NOPs for these are in cross-misc.
 
@@ -271,9 +344,6 @@ error if any of PACKAGES is not a valid package designator."
 ;;; WITH-SINGLE-PACKAGE-LOCKED-ERROR
 (defun assert-package-unlocked (package &optional format-control
                                 &rest format-arguments)
-  #!-sb-package-locks
-  (declare (ignore format-control format-arguments))
-  #!+sb-package-locks
   (when (package-lock-violation-p package)
     (package-lock-violation package
                             :format-control format-control
@@ -288,9 +358,6 @@ error if any of PACKAGES is not a valid package designator."
 ;;; this.
 (defun assert-symbol-home-package-unlocked (name &optional format-control
                                             &rest format-arguments)
-  #!-sb-package-locks
-  (declare (ignore format-control format-arguments))
-  #!+sb-package-locks
   (let* ((symbol (etypecase name
                    (symbol name)
                    ;; Istm that the right way to declare that you want to allow
@@ -543,9 +610,11 @@ REMOVE-PACKAGE-LOCAL-NICKNAME, and the DEFPACKAGE option :LOCAL-NICKNAMES."
 
 ;;; Perform (GETHASH NAME TABLE) and then unwrap the value if it is a list.
 ;;; List vs nonlist disambiguates a nickname from the primary name.
+;;; And never return the symbol :DELETED.
 (defun %get-package (name table)
-  (let ((found (gethash name table)))
-    (if (listp found) (car found) found)))
+  (let ((found (info-gethash name table)))
+    (cond ((listp found) (car found))
+          ((neq found :deleted) found))))
 
 ;;; This is undocumented and unexported for now, but the idea is that by
 ;;; making this a generic function then packages with custom package classes
@@ -626,10 +695,49 @@ REMOVE-PACKAGE-LOCAL-NICKNAME, and the DEFPACKAGE option :LOCAL-NICKNAMES."
                (decf (package-hashtable-deleted table))))) ; tombstone
       (declare (fixnum i)))))
 
+;;; Insert a mapping from NAME (a string) to OBJECT (a package or singleton
+;;; list of a package) into TABLE (the hashtable in *PACKAGE-NAMES*),
+;;; taking care to adjust the count of phantom entries.
+(defun %register-package (table name object)
+  (let ((oldval (info-gethash name table)))
+    (unless oldval ; if any value existed, no new physical cell is claimed
+      (when (> (info-env-tombstones table)
+               (floor (info-storage-capacity (info-env-storage table)) 4))
+        ;; Otherwise, when >1/4th of the table consists of tombstones,
+        ;; then rebuild the table.
+        (%rebuild-package-names table)
+        (when (eq oldval :deleted)
+          (setq oldval nil))))
+    (setf (info-gethash name table) object)
+    (when (eq oldval :deleted)
+      (decf (info-env-tombstones table)))))
+
+;;; Rebuild the *PACKAGE-NAMES* table.
+;;; The calling thread must own the mutex on *PACKAGE-NAMES* so that
+;;; this is synchronized across all insertion and deletion operations.
+(defun %rebuild-package-names (old)
+  (let ((new (copy-structure old))
+        (nondeleted-count (- (info-env-count old)
+                             (info-env-tombstones old))))
+    (setf (info-env-storage new) (make-info-storage nondeleted-count)
+          (info-env-count new) 0
+          (info-env-tombstones new) 0)
+    (info-maphash (lambda (key value)
+                    (unless (eq value :deleted)
+                      (setf (info-gethash key new) value)))
+                  old)
+    (setf (info-env-storage old) (info-env-storage new)
+          (info-env-count old) (info-env-count new)
+          (info-env-tombstones old) 0)
+    old))
+
 ;;; Resize the package hashtables of all packages so that their load
 ;;; factor is *PACKAGE-HASHTABLE-IMAGE-LOAD-FACTOR*. Called from
 ;;; SAVE-LISP-AND-DIE to optimize space usage in the image.
 (defun tune-hashtable-sizes-of-all-packages ()
+  (with-package-names (table)
+    (when (plusp (info-env-tombstones table))
+      (%rebuild-package-names table)))
   (flet ((tune-table-size (table)
            (resize-package-hashtable
             table
@@ -701,30 +809,30 @@ REMOVE-PACKAGE-LOCAL-NICKNAME, and the DEFPACKAGE option :LOCAL-NICKNAMES."
       (resize-package-hashtable table (* used 2)))))
 
 ;;; Enter any new NICKNAMES for PACKAGE into *PACKAGE-NAMES*. If there is a
-;;; conflict then give the user a chance to do something about it. Caller is
-;;; responsible for having acquired the mutex via WITH-PACKAGES.
+;;; conflict then give the user a chance to do something about it.
+;;; Package names do not affect the uses/used-by relation,
+;;; so this can be done without the package graph lock held.
 (defun %enter-new-nicknames (package nicknames &aux (val (list package)))
   (declare (type list nicknames))
-  (dolist (n nicknames)
-    (let ((found (with-package-names (names)
-                    (or (%get-package (the simple-string n) names)
-                        (progn
-                          (setf (gethash n names) val)
-                          (push n (package-%nicknames package))
-                          package)))))
+  (dolist (nickname nicknames)
+    (let ((found (or (%get-package (the simple-string nickname) *package-names*)
+                     (with-package-names (table)
+                       (%register-package table nickname val)
+                       (push nickname (package-%nicknames package))
+                       package))))
       (cond ((eq found package))
-            ((string= (the string (package-%name found)) n)
+            ((string= (the string (package-%name found)) nickname)
              (signal-package-cerror
               package
               "Ignore this nickname."
               "~S is a package name, so it cannot be a nickname for ~S."
-              n (package-%name package)))
+              nickname (package-%name package)))
             (t
              (signal-package-cerror
               package
               "Leave this nickname alone."
               "~S is already a nickname for ~S."
-              n (package-%name found)))))))
+              nickname (package-%name found)))))))
 
 ;;; ANSI specifies that:
 ;;;  (1) MAKE-PACKAGE and DEFPACKAGE use the same default package-use-list
@@ -733,7 +841,15 @@ REMOVE-PACKAGE-LOCAL-NICKNAME, and the DEFPACKAGE option :LOCAL-NICKNAMES."
 ;;; For OAOO reasons we give a name to this value and then use #. readmacro
 ;;; to splice it in as a constant. Anyone who actually wants a random value
 ;;; is free to :USE (PACKAGE-USE-LIST :CL-USER) or whatever.
-(defglobal *!default-package-use-list* nil)
+;;;
+;;; This must not use DEFGLOBAL, which calls SET-SYMBOL-GLOBAL-VALUE at
+;;; compile-time. The xc stub for that is defined to error out on purpose,
+;;; since there is no such thing as a global variable in portable CL.
+;;; Anyway this is a trivial compile-time constant that disappears after
+;;; self-build, so efficiency of reads (one of the concerns of DEFGLOBAL,
+;;; the other being documentation of intent) doesn't matter in the least.
+(eval-when (:compile-toplevel :load-toplevel :execute)
+  (defvar *!default-package-use-list* nil))
 
 (defun make-package (name &key
                           (use '#.*!default-package-use-list*)
@@ -779,9 +895,35 @@ implementation it is ~S." *!default-package-use-list*)
          ;; USE-PACKAGE, but I need to check what kinds of errors can be caused by
          ;; USE-PACKAGE, too.
          (%enter-new-nicknames package nicks)
-         (return (setf (gethash name *package-names*) package))))
+         ;; The name table is actually multi-writer concurrent, but due to
+         ;; lazy removal of :DELETED entries we want to enforce a single-writer.
+         ;; We're inside WITH-PACKAGE-GRAPH so this is already synchronized with
+         ;; other MAKE-PACKAGE operations, but we need the additional lock
+         ;; so that it synchronizes with RENAME-PACKAGE.
+         (with-package-names (table)
+           (%register-package table name package))
+         (return package)))
      (bug "never")))
 
+(flet ((remove-names (package name-table)
+         ;; An INFO-HASHTABLE does not support REMHASH. We can simulate it
+         ;; by changing the value to :DELETED.
+         ;; (NIL would be preferable, but INFO-GETHASH does not return
+         ;; a secondary value indicating whether the NIL was by default
+         ;; or found, not does it take a different default to return).
+         ;; At some point the table might contain more deleted values than
+         ;; useful values. We call %REBUILD-PACKAGE-NAMES to rectify that.
+         (dx-let ((names (cons (package-name package)
+                               (package-nicknames package)))
+                  (i 0))
+           (dolist (name names)
+             ;; Aver that the following SETF doesn't insert a new <k,v> pair.
+             (aver (info-gethash name name-table))
+             (setf (info-gethash name name-table) :deleted)
+             (incf i))
+           (incf (info-env-tombstones name-table) i))
+         nil))
+
 ;;; Change the name if we can, blast any old nicknames and then
 ;;; add in any new ones.
 ;;;
@@ -810,16 +952,14 @@ implementation it is ~S." *!default-package-use-list*)
            (assert-package-unlocked
             package "renaming as ~A~@[ with nickname~*~P ~1@*~{~A~^, ~}~]"
             name nicks (length nicks)))
-         (with-package-names (names)
+         (with-package-names (table)
            ;; Check for race conditions now that we have the lock.
            (unless (eq package (find-package package-designator))
              (go :restart))
            ;; Do the renaming.
-           (remhash (package-%name package) names)
-           (dolist (n (package-%nicknames package))
-             (remhash n names))
+           (remove-names package table)
+           (%register-package table name package)
            (setf (package-%name package) name
-                 (gethash name names) package
                  (package-%nicknames package) ()))
          (%enter-new-nicknames package nicks))
        (return package))))
@@ -853,7 +993,6 @@ implementation it is ~S." *!default-package-use-list*)
                      (mapcar #'package-name use-list))
                     (dolist (p use-list)
                       (unuse-package package p))))
-                #!+sb-package-locks
                 (dolist (p (package-implements-list package))
                   (remove-implementation-package package p))
                 (with-package-graph ()
@@ -877,10 +1016,8 @@ implementation it is ~S." *!default-package-use-list*)
                   ;; many smaller tables for no good reason.
                   (do-symbols (sym package)
                     (unintern sym package))
-                  (with-package-names (names)
-                    (remhash (package-name package) names)
-                    (dolist (nick (package-nicknames package))
-                      (remhash nick names))
+                  (with-package-names (table)
+                    (remove-names package table)
                     (setf (package-%name package) nil
                           ;; Setting PACKAGE-%NAME to NIL is required in order to
                           ;; make PACKAGE-NAME return NIL for a deleted package as
@@ -896,17 +1033,24 @@ implementation it is ~S." *!default-package-use-list*)
                         (package-external-symbols package)
                         (make-package-hashtable 0)))
                 (return-from delete-package t)))))))
+) ; end FLET
+
+(defmacro !do-packages ((package) &body body)
+  ;; INFO-MAPHASH is not intrinsically threadsafe - but actually
+  ;; quite easy to fix - so meanwhile until it's fixed, grab the lock.
+  `(with-package-names (.table.)
+      (info-maphash
+       (lambda (.name. ,package)
+         (declare (ignore .name.))
+         (unless (or (listp ,package) (eq ,package :deleted))
+           ,@body))
+       .table.)))
 
 (defun list-all-packages ()
   "Return a list of all existing packages."
-  (let ((res ()))
-    (with-package-names (names)
-      (maphash (lambda (k v)
-                 (declare (ignore k))
-                 (unless (listp v)
-                   (push v res)))
-               names))
-    res))
+  (let ((result ()))
+    (!do-packages (package) (push package result))
+    result))
 
 (macrolet ((find/intern (function &rest more-args)
              ;; Both %FIND-SYMBOL and %INTERN require a SIMPLE-STRING,
@@ -1039,12 +1183,12 @@ implementation it is ~S." *!default-package-use-list*)
   ((function :initarg :function :reader name-conflict-function)
    (datum :initarg :datum :reader name-conflict-datum)
    (symbols :initarg :symbols :reader name-conflict-symbols))
-  (:default-initargs :references (list '(:ansi-cl :section (11 1 1 2 5))))
+  (:default-initargs :references '((:ansi-cl :section (11 1 1 2 5))))
   (:report
    (lambda (c s)
      (format s "~@<~S ~S causes name-conflicts in ~S between the ~
                 following symbols: ~2I~@:_~
-                ~{~/sb-impl::print-symbol-with-prefix/~^, ~}~:@>"
+                ~{~/sb-ext:print-symbol-with-prefix/~^, ~}~:@>"
              (name-conflict-function c)
              (name-conflict-datum c)
              (package-error-package c)
@@ -1114,7 +1258,7 @@ implementation it is ~S." *!default-package-use-list*)
                    (*print-pretty* t))
               (format *query-io* "~&~@<Select a symbol to be made accessible in ~
                               package ~A:~2I~@:_~{~{~V,' D. ~
-                              ~/sb-impl::print-symbol-with-prefix/~}~@:_~}~
+                              ~/sb-ext:print-symbol-with-prefix/~}~@:_~}~
                               ~@:>"
                       (package-name package)
                       (loop for s in symbols
@@ -1163,10 +1307,8 @@ uninterned."
     (let* ((package (find-undeleted-package-or-lose package))
            (name (symbol-name symbol))
            (shadowing-symbols (package-%shadowing-symbols package)))
-      (declare (list shadowing-symbols))
-
       (with-single-package-locked-error ()
-        (when (find-symbol name package)
+        (when (nth-value 1 (find-symbol name package))
           (assert-package-unlocked package "uninterning ~A" name))
 
         ;; If a name conflict is revealed, give us a chance to
@@ -1470,15 +1612,11 @@ PACKAGE."
 (defun find-all-symbols (string-or-symbol)
   "Return a list of all symbols in the system having the specified name."
   (let ((string (string string-or-symbol))
-        (res ()))
-    (with-package-names (names)
-      (maphash (lambda (k v)
-                 (declare (ignore k))
-                 (unless (listp v) ; ignore nickname entries
-                   (multiple-value-bind (s w) (find-symbol string v)
-                     (when w (pushnew s res)))))
-               names))
-    res))
+        (result ()))
+    (!do-packages (package)
+      (multiple-value-bind (symbol found) (find-symbol string package)
+        (when found (pushnew symbol result))))
+    result))
 
 ;;;; APROPOS and APROPOS-LIST
 
@@ -1539,7 +1677,10 @@ PACKAGE."
 
 (defun !package-cold-init ()
   (setf *package-graph-lock* (sb!thread:make-mutex :name "Package Graph Lock")
-        *package-names* (make-hash-table :test 'equal :synchronized t))
+        *package-names*
+        (make-info-hashtable
+         :comparator (lambda (a b) (and (not (eql a 0)) (string= a b)))
+         :hash-function #'sxhash))
   (with-package-names (names)
     (dolist (spec *!initial-symbols*)
       (let ((pkg (car spec)) (symbols (cdr spec)))
@@ -1556,10 +1697,9 @@ PACKAGE."
         (setf (package-%local-nicknames pkg) nil
               (package-%locally-nicknamed-by pkg) nil
               (package-source-location pkg) nil
-              (gethash (package-%name pkg) names) pkg)
+              (info-gethash (package-%name pkg) names) pkg)
         (dolist (nick (package-%nicknames pkg))
-          (setf (gethash nick names) (list pkg)))
-        #!+sb-package-locks
+          (setf (info-gethash nick names) (list pkg)))
         (setf (package-lock pkg) nil
               (package-%implementation-packages pkg) nil))))
 
@@ -1657,9 +1797,6 @@ PACKAGE."
               (scan))))))
 
 (defun program-assert-symbol-home-package-unlocked (context symbol control)
-  #!-sb-package-locks
-  (declare (ignore context symbol control))
-  #!+sb-package-locks
   (handler-bind ((package-lock-violation
                   (lambda (condition)
                     (ecase context
