@@ -157,11 +157,6 @@ exited. The offending thread can be accessed using THREAD-ERROR-THREAD."))
 ;;; gencgc and numbers on the stack (returned by GET-LISP-OBJ-ADDRESS)
 ;;; are treated as references.
 
-;;; set the doc here because in early-thread FDOCUMENTATION is not
-;;; available, yet
-(setf (fdocumentation '*current-thread* 'variable)
-      "Bound in each thread to the thread itself.")
-
 (setf
  (fdocumentation 'thread-name 'function)
  "Name of the thread. Can be assigned to using SETF. Thread names can be
@@ -215,9 +210,12 @@ potentially stale even before the function returns, as the thread may exit at
 any time."
   (thread-%alive-p thread))
 
+;; NB: ephemeral threads must terminate strictly before the test of NTHREADS>1
+;; in DEINIT, i.e. this is not a promise that the thread will terminate
+;; just-in-time for the final call out to save, but rather by an earlier time.
 (defun thread-ephemeral-p (thread)
   "Return T if THREAD is `ephemeral', which indicates that this thread is
-used by SBCL for internal purposes, and specifically that it knows how to
+used by SBCL for internal purposes, and specifically that our runtime knows how
 to terminate this thread cleanly prior to core file saving without signalling
 an error in that case."
   (thread-%ephemeral-p thread))
@@ -270,7 +268,10 @@ created and old ones may exit at any time."
 
 (declaim (inline current-thread-sap))
 (defun current-thread-sap ()
-  (sb!vm::current-thread-offset-sap sb!vm::thread-this-slot))
+  #!+sb-thread
+  (sb!vm::current-thread-offset-sap sb!vm::thread-this-slot)
+  #!-sb-thread
+  (int-sap 0))
 
 (declaim (inline current-thread-os-thread))
 (defun current-thread-os-thread ()
@@ -286,16 +287,17 @@ created and old ones may exit at any time."
   (/show0 "Entering INIT-INITIAL-THREAD")
   (setf sb!impl::*exit-lock* (make-mutex :name "Exit Lock")
         *make-thread-lock* (make-mutex :name "Make-Thread Lock"))
-  (let ((initial-thread (%make-thread :name "main thread"
-                                      :%alive-p t)))
-    (setf (thread-os-thread initial-thread) (current-thread-os-thread)
-          *initial-thread* initial-thread
-          *current-thread* initial-thread)
+  (let ((thread (%make-thread :name "main thread"
+                              :%alive-p t)))
+    (setf (thread-os-thread thread) (current-thread-os-thread)
+          (thread-primitive-thread thread) (sap-int (current-thread-sap))
+          *initial-thread* thread
+          *current-thread* thread)
     (grab-mutex (thread-result-lock *initial-thread*))
     ;; Either *all-threads* is empty or it contains exactly one thread
     ;; in case we are in reinit since saving core with multiple
     ;; threads doesn't work.
-    (setq *all-threads* (list initial-thread))))
+    (setq *all-threads* (list thread))))
 
 (defun main-thread ()
   "Returns the main thread of the process."
@@ -366,12 +368,12 @@ See also: RETURN-FROM-THREAD and SB-EXT:EXIT."
 ;;;; Aliens, low level stuff
 
 (define-alien-routine "kill_safely"
-    integer
+    int
   (os-thread #!-alpha unsigned #!+alpha unsigned-int)
   (signal int))
 
 (define-alien-routine "wake_thread"
-    integer
+    int
   (os-thread unsigned))
 
 #!+sb-thread
@@ -412,9 +414,7 @@ See also: RETURN-FROM-THREAD and SB-EXT:EXIT."
     `(let* ((,n-thread ,thread)
             (,n-lock ,lock)
             (,n-timeout ,(when timeoutp
-                           `(or ,timeout
-                                (when sb!impl::*deadline*
-                                  sb!impl::*deadline-seconds*))))
+                           `(or ,timeout sb!impl::*deadline*)))
             (,new (if ,n-timeout
                       ;; Using CONS tells the rest of the system there's a
                       ;; timeout in place, so it isn't considered a deadlock.
@@ -461,7 +461,7 @@ HOLDING-MUTEX-P."
   ;; Make sure to get the current value.
   (sb!ext:compare-and-swap (mutex-%owner mutex) nil nil))
 
-(sb!ext:defglobal **deadlock-lock** nil)
+(sb!ext:define-load-time-global **deadlock-lock** nil)
 
 #!+(or (not sb-thread) sb-futex)
 (defstruct (waitqueue (:copier nil) (:constructor make-waitqueue (&key name)))
@@ -550,6 +550,112 @@ HOLDING-MUTEX-P."
         (detect-deadlock origin)
         t))))
 
+;;;; WAIT-FOR -- waiting on arbitrary conditions
+
+(defun %%wait-for (test stop-sec stop-usec)
+  (declare (function test))
+  (labels ((try ()
+             (declare (optimize (safety 0)))
+             (awhen (funcall test)
+               (return-from %%wait-for it)))
+           (tick (sec usec)
+             (declare (type fixnum sec usec))
+             ;; TICK is microseconds
+             (+ usec (* 1000000 sec)))
+           (get-tick ()
+             (multiple-value-call #'tick
+               (decode-internal-time (get-internal-real-time)))))
+    (let* ((timeout-tick (when stop-sec (tick stop-sec stop-usec)))
+           (start (get-tick))
+           ;; Rough estimate of how long a single attempt takes.
+           (try-ticks (progn
+                        (try) (try) (try)
+                        (max 1 (truncate (- (get-tick) start) 3)))))
+      ;; Scale sleeping between attempts:
+      ;;
+      ;; Start by sleeping for as many ticks as an average attempt
+      ;; takes, then doubling for each attempt.
+      ;;
+      ;; Max out at 0.1 seconds, or the 2 x time of a single try,
+      ;; whichever is longer -- with a hard cap of 10 seconds.
+      ;;
+      ;; FIXME: Maybe the API should have a :MAX-SLEEP argument?
+      (loop with max-ticks = (max 100000 (min (* 2 try-ticks)
+                                              (expt 10 7)))
+            for scale of-type fixnum = 1
+            then (let ((x (logand most-positive-fixnum (* 2 scale))))
+                   (if (> scale x)
+                       most-positive-fixnum
+                       x))
+            do (try)
+               (let* ((now (get-tick))
+                      (sleep-ticks (min (* try-ticks scale) max-ticks))
+                      (sleep
+                        (if timeout-tick
+                            ;; If sleep would take us past the
+                            ;; timeout, shorten it so it's just
+                            ;; right.
+                            (if (>= (+ now sleep-ticks) timeout-tick)
+                                (- timeout-tick now)
+                                sleep-ticks)
+                            sleep-ticks)))
+                 (declare (type fixnum sleep))
+                 (cond ((plusp sleep)
+                        ;; microseconds to seconds and nanoseconds
+                        (multiple-value-bind (sec nsec)
+                            (truncate (* 1000 sleep) (expt 10 9))
+                          (with-interrupts
+                            (sb!unix:nanosleep sec nsec))))
+                       (t
+                        (return-from %%wait-for nil))))))))
+
+(defun %wait-for (test timeout)
+  (declare (function test))
+  (tagbody
+   :restart
+     (multiple-value-bind (to-sec to-usec stop-sec stop-usec deadlinep)
+         (decode-timeout timeout)
+       (declare (ignore to-sec to-usec))
+       (return-from %wait-for
+         (or (%%wait-for test stop-sec stop-usec)
+             (when deadlinep
+               (signal-deadline)
+               (go :restart)))))))
+
+(defmacro sb!ext:wait-for (test-form &key timeout)
+  "Wait until TEST-FORM evaluates to true, then return its primary value.
+If TIMEOUT is provided, waits at most approximately TIMEOUT seconds before
+returning NIL.
+
+If WITH-DEADLINE has been used to provide a global deadline, signals a
+DEADLINE-TIMEOUT if TEST-FORM doesn't evaluate to true before the
+deadline.
+
+Experimental: subject to change without prior notice."
+  `(dx-flet ((wait-for-test () (progn ,test-form)))
+     (%wait-for #'wait-for-test ,timeout)))
+
+(defmacro with-progressive-timeout ((name &key seconds)
+                                    &body body)
+  "Binds NAME as a local function for BODY. Each time #'NAME is called, it
+returns SECONDS minus the time that has elapsed since BODY was entered, or
+zero if more time than SECONDS has elapsed. If SECONDS is NIL, #'NAME
+returns NIL each time."
+  (with-unique-names (deadline time-left sec)
+    `(let* ((,sec ,seconds)
+            (,deadline
+              (when ,sec
+                (+ (get-internal-real-time)
+                   (round (* ,seconds internal-time-units-per-second))))))
+       (flet ((,name ()
+                (when ,deadline
+                  (let ((,time-left (- ,deadline (get-internal-real-time))))
+                    (if (plusp ,time-left)
+                        (* (coerce ,time-left 'single-float)
+                           (load-time-value (/ 1.0f0 internal-time-units-per-second) t))
+                        0)))))
+         ,@body))))
+
 (defun %try-mutex (mutex new-owner)
   (declare (type mutex mutex) (optimize (speed 3)))
   (barrier (:read))
@@ -576,6 +682,7 @@ HOLDING-MUTEX-P."
 #!+sb-thread
 (defun %%wait-for-mutex (mutex new-owner to-sec to-usec stop-sec stop-usec)
   (declare (type mutex mutex) (optimize (speed 3)))
+  (declare (sb!ext:muffle-conditions sb!ext:compiler-note))
   #!-sb-futex
   (declare (ignore to-sec to-usec))
   #!-sb-futex
@@ -593,7 +700,7 @@ HOLDING-MUTEX-P."
            ;; Check for pending interrupts.
            (with-interrupts nil)))
     (declare (dynamic-extent #'cas))
-    (sb!impl::%%wait-for #'cas stop-sec stop-usec))
+    (%%wait-for #'cas stop-sec stop-usec))
   #!+sb-futex
   ;; This is a fairly direct translation of the Mutex 2 algorithm from
   ;; "Futexes are Tricky" by Ulrich Drepper.
@@ -822,7 +929,7 @@ IF-NOT-OWNER is :FORCE)."
   #!+sb-thread
   (let ((me *current-thread*))
     (barrier (:read))
-    (assert (eq me (mutex-%owner mutex)))
+    (aver (eq me (mutex-%owner mutex)))
     (let ((status :interrupted))
       ;; Need to disable interrupts so that we don't miss grabbing
       ;; the mutex on our way out.
@@ -841,7 +948,7 @@ IF-NOT-OWNER is :FORCE)."
                                       :ok)))
                              (declare (dynamic-extent #'wakeup))
                              (allow-with-interrupts
-                               (sb!impl::%%wait-for #'wakeup stop-sec stop-usec)))
+                               (%%wait-for #'wakeup stop-sec stop-usec)))
                            :timeout)))
                #!+sb-futex
                (with-pinned-objects (queue me)
@@ -1020,11 +1127,11 @@ must be held by this thread during this call."
   "Semaphore type. The fact that a SEMAPHORE is a STRUCTURE-OBJECT
 should be considered an implementation detail, and may change in the
 future."
-  (name    nil :type (or null thread-name))
+  (name    nil :type (or null thread-name) :read-only t)
   (%count    0 :type (integer 0))
   (waitcount 0 :type sb!vm:word)
-  (mutex (make-mutex))
-  (queue (make-waitqueue)))
+  (mutex (make-mutex :name "semaphore lock") :read-only t)
+  (queue (make-waitqueue) :read-only t))
 
 (setf (fdocumentation 'semaphore-name 'function)
       "The name of the semaphore INSTANCE. Setfable."
@@ -1231,7 +1338,7 @@ on this semaphore, then N of them is woken up."
 
 (defmacro with-new-session (args &body forms)
   (declare (ignore args))               ;for extensibility
-  (sb!int:with-unique-names (fb-name)
+  (with-unique-names (fb-name)
     `(labels ((,fb-name () ,@forms))
       (call-with-new-session (function ,fb-name)))))
 
@@ -1244,11 +1351,23 @@ on this semaphore, then N of them is woken up."
   ;; Lisp-side cleanup
   (with-all-threads-lock
     (setf (thread-%alive-p thread) nil)
-    (setf (thread-os-thread thread) 0)
+    (setf (thread-os-thread thread)
+          (ldb (byte sb!vm:n-word-bits 0) -1))
     (setq *all-threads* (delq thread *all-threads*))
     (when *session*
       (%delete-thread-from-session thread *session*))))
 
+(defvar sb!ext:*invoke-debugger-hook* nil
+  "This is either NIL or a designator for a function of two arguments,
+   to be run when the debugger is about to be entered.  The function is
+   run with *INVOKE-DEBUGGER-HOOK* bound to NIL to minimize recursive
+   errors, and receives as arguments the condition that triggered
+   debugger entry and the previous value of *INVOKE-DEBUGGER-HOOK*
+
+   This mechanism is an SBCL extension similar to the standard *DEBUGGER-HOOK*.
+   In contrast to *DEBUGGER-HOOK*, it is observed by INVOKE-DEBUGGER even when
+   called by BREAK.")
+
 (defun %exit-other-threads ()
   ;; Grabbing this lock prevents new threads from
   ;; being spawned, and guarantees that *ALL-THREADS*
@@ -1260,6 +1379,12 @@ on this semaphore, then N of them is woken up."
           (current *current-thread*)
           (joinees nil)
           (main nil))
+      ;; Don't invoke the debugger on errors in cleanup forms in unwind-protect
+      (setf sb!ext:*invoke-debugger-hook*
+            (lambda (c h)
+              (sb!debug::debugger-disabled-hook c h :quit nil)
+              (abort-thread :allow-exit t)))
+      (sb!impl::finalizer-thread-stop)
       (dolist (thread (list-all-threads))
         (cond ((eq thread current))
               ((main-thread-p thread)
@@ -1375,7 +1500,7 @@ session."
   (first (interactive-threads session)))
 
 (defun make-listener-thread (tty-name)
-  (assert (probe-file tty-name))
+  (aver (probe-file tty-name))
   (let* ((in (sb!unix:unix-open (namestring tty-name) sb!unix:o_rdwr #o666))
          (out (sb!unix:unix-dup in))
          (err (sb!unix:unix-dup in)))
@@ -1398,113 +1523,80 @@ session."
                  (with-new-session ()
                    (unwind-protect
                         (sb!impl::toplevel-repl nil)
-                     (sb!int:flush-standard-output-streams))))))
+                     (flush-standard-output-streams))))))
       (make-thread #'thread-repl))))
 
 
 ;;;; The beef
 
 #!+sb-thread
-(defun initial-thread-function-trampoline
-    (thread setup-sem real-function thread-list arguments arg1 arg2 arg3)
-  ;; In time we'll move some of the binding presently done in C here
-  ;; too.
-  ;;
-  ;; KLUDGE: Here we have a magic list of variables that are not
-  ;; thread-safe for one reason or another.  As people report problems
-  ;; with the thread safety of certain variables, (e.g. "*print-case* in
-  ;; multiple threads broken", sbcl-devel 2006-07-14), we add a few more
-  ;; bindings here.  The Right Thing is probably some variant of
-  ;; Allegro's *cl-default-special-bindings*, as that is at least
-  ;; accessible to users to secure their own libraries.
-  ;;   --njf, 2006-07-15
-  ;;
-  ;; FIXME: because #<"main thread"> does not come through here, the global
-  ;; values of these symbols matter. That's bad! It would be nicer for codegen
-  ;; if these symbols (and some others) never required testing for
-  ;; NO-TLS-VALUE-MARKER and potential fallback to loading symbol-global-value.
-  ;;
-  ;; As it is, this lambda must not cons until we are ready to run
-  ;; GC. Be very careful.
-  (let* ((*current-thread* thread)
-         (*restart-clusters* nil)
-         (*handler-clusters* sb!kernel::**initial-handler-clusters**)
-         (*exit-in-process* nil)
-         (sb!impl::*deadline* nil)
-         (sb!impl::*deadline-seconds* nil)
-         (sb!impl::*step-out* nil)
-         ;; internal reader variables
-         (sb!impl::*token-buf-pool* nil)
-         (sb!impl::*ignored-package-locks* :invalid)
-         (sb!impl::*descriptor-handlers* nil)) ; serve-event
-    ;; *ALLOC-SIGNAL* is made thread-local by create_thread_struct()
-    ;; so this assigns into TLS, not the global value.
-    (setf sb!vm:*alloc-signal* *default-alloc-signal*)
-    (setf (thread-os-thread thread) (current-thread-os-thread))
-    (with-mutex ((thread-result-lock thread))
-      ;; Normally the list is allocated in the parent thread to
-      ;; avoid consing in a nascent thread.
+(defun initial-thread-function-trampoline (thread setup-sem real-function arguments)
+  ;; Can't initiate GC before *current-thread* is set, otherwise the
+  ;; locks grabbed by SUB-GC wouldn't function.
+  ;; Other threads can GC with impunity.
+  (setf *current-thread* thread ; is thread-local already
+        (thread-os-thread thread) (current-thread-os-thread)
+        (thread-primitive-thread thread) (sap-int (current-thread-sap)))
+  ;; *ALLOC-SIGNAL* is made thread-local by create_thread_struct()
+  ;; so this assigns into TLS, not the global value.
+  (setf sb!vm:*alloc-signal* *default-alloc-signal*)
+  (with-mutex ((thread-result-lock thread))
+    (let* ((thread-list (list thread thread))
+           (session-cons (cdr thread-list)))
+      (with-all-threads-lock
+        (setf (cdr thread-list) *all-threads*
+              *all-threads* thread-list))
+      (let ((session *session*))
+        (with-session-lock (session)
+          (setf (cdr session-cons) (session-threads session)
+                (session-threads session) session-cons))))
+    (setf (thread-%alive-p thread) t)
 
-      (let* ((thread-list (or thread-list
-                              (list thread thread)))
-             (session-cons (cdr thread-list)))
-        (with-all-threads-lock
-          (setf (cdr thread-list) *all-threads*
-                *all-threads* thread-list))
-        (let ((session *session*))
-          (with-session-lock (session)
-            (setf (cdr session-cons) (session-threads session)
-                  (session-threads session) session-cons))))
-      (setf (thread-%alive-p thread) t)
+    (when setup-sem
+      (signal-semaphore setup-sem)
+      ;; setup-sem was dx-allocated, set it to NIL so that the
+      ;; backtrace doesn't get confused
+      (setf setup-sem nil))
 
-      (when setup-sem
-        (signal-semaphore setup-sem)
-        ;; setup-sem was dx-allocated, set it to NIL so that the
-        ;; backtrace doesn't get confused
-        (setf setup-sem nil))
-
-      ;; Using handling-end-of-the-world would be a bit tricky
-      ;; due to other catches and interrupts, so we essentially
-      ;; re-implement it here. Once and only once more.
-      (catch 'sb!impl::toplevel-catcher
-        (catch 'sb!impl::%end-of-the-world
-          (catch '%abort-thread
-            (restart-bind ((abort
-                             (lambda ()
-                               (throw '%abort-thread nil))
-                             :report-function
-                             (lambda (stream)
-                               (format stream "~@<abort thread (~a)~@:>"
-                                       *current-thread*))))
-              (without-interrupts
-                (unwind-protect
-                     (with-local-interrupts
-                       (setf *gc-inhibit* nil) ;for foreign callbacks
-                       (sb!unix::unblock-deferrable-signals)
-                       (setf (thread-result thread)
-                             (prog1
-                                 (multiple-value-list
-                                  (unwind-protect
-                                       (catch '%return-from-thread
-                                         (if (listp arguments)
-                                             (apply real-function arguments)
-                                             (funcall real-function arg1 arg2 arg3)))
-                                    (when *exit-in-process*
-                                      (sb!impl::call-exit-hooks))))
-                               #!+sb-safepoint
-                               (sb!kernel::gc-safepoint))))
-                  ;; we're going down, can't handle interrupts
-                  ;; sanely anymore. gc remains enabled.
-                  (block-deferrable-signals)
-                  ;; we don't want to run interrupts in a dead
-                  ;; thread when we leave without-interrupts.
-                  ;; this potentially causes important
-                  ;; interupts to be lost: sigint comes to
-                  ;; mind.
-                  (setq *interrupt-pending* nil)
-                  #!+sb-thruption
-                  (setq *thruption-pending* nil)
-                  (handle-thread-exit thread)))))))))
+    ;; Using handling-end-of-the-world would be a bit tricky
+    ;; due to other catches and interrupts, so we essentially
+    ;; re-implement it here. Once and only once more.
+    (catch 'sb!impl::toplevel-catcher
+      (catch 'sb!impl::%end-of-the-world
+        (catch '%abort-thread
+          (restart-bind ((abort
+                           (lambda ()
+                             (throw '%abort-thread nil))
+                           :report-function
+                           (lambda (stream)
+                             (format stream "~@<abort thread (~a)~@:>"
+                                     *current-thread*))))
+            (without-interrupts
+              (unwind-protect
+                   (with-local-interrupts
+                     (sb!unix::unblock-deferrable-signals)
+                     (setf (thread-result thread)
+                           (prog1
+                               (multiple-value-list
+                                (unwind-protect
+                                     (catch '%return-from-thread
+                                       (apply real-function arguments))
+                                  (when *exit-in-process*
+                                    (sb!impl::call-exit-hooks))))
+                             #!+sb-safepoint
+                             (sb!kernel::gc-safepoint))))
+                ;; we're going down, can't handle interrupts
+                ;; sanely anymore. gc remains enabled.
+                (block-deferrable-signals)
+                ;; we don't want to run interrupts in a dead
+                ;; thread when we leave without-interrupts.
+                ;; this potentially causes important
+                ;; interupts to be lost: sigint comes to
+                ;; mind.
+                (setq *interrupt-pending* nil)
+                #!+sb-thruption
+                (setq *thruption-pending* nil)
+                (handle-thread-exit thread))))))))
   (values))
 
 (defun make-thread (function &key name arguments ephemeral)
@@ -1533,27 +1625,24 @@ See also: RETURN-FROM-THREAD, ABORT-THREAD."
            (real-function (coerce function 'function))
            (arguments     (ensure-list arguments))
            #!+(or win32 darwin)
-           (fp-modes (dpb 0 sb!vm::float-sticky-bits ;; clear accrued bits
-                          (sb!vm:floating-point-modes)))
-           ;; Allocate in the parent
-           (thread-list (list thread thread)))
+           (fp-modes (dpb 0 sb!vm:float-sticky-bits ;; clear accrued bits
+                          (sb!vm:floating-point-modes))))
       (declare (dynamic-extent setup-sem))
       (dx-flet ((initial-thread-function ()
                   ;; Inherit parent thread's FP modes
                   #!+(or win32 darwin)
                   (setf (sb!vm:floating-point-modes) fp-modes)
                   ;; As it is, this lambda must not cons until we are
-                  ;; ready to run GC. Be very careful.
-                  (initial-thread-function-trampoline
-                   thread setup-sem real-function thread-list
-                   arguments nil nil nil)))
-        ;; If the starting thread is stopped for gc before it signals
-        ;; the semaphore then we'd be stuck.
-        (assert (not *gc-inhibit*))
+                  ;; ready to run GC. Be careful.
+                  (initial-thread-function-trampoline thread setup-sem
+                                                      real-function arguments)))
+        ;; Holding mutexes or waiting on sempahores inside WITHOUT-GCING will lock up
+        (aver (not *gc-inhibit*))
         ;; Keep INITIAL-FUNCTION in the dynamic extent until the child
         ;; thread is initialized properly. Wrap the whole thing in
-        ;; WITHOUT-INTERRUPTS because we pass INITIAL-FUNCTION to
-        ;; another thread.
+        ;; WITHOUT-INTERRUPTS (via WITH-SYSTEM-MUTEX) because we pass
+        ;; INITIAL-FUNCTION to another thread.
+        ;; (Does WITHOUT-INTERRUPTS really matter now that it's DXed?)
         (with-system-mutex (*make-thread-lock*)
           (if (zerop
                (%create-thread (get-lisp-obj-address #'initial-thread-function)))
@@ -1619,10 +1708,14 @@ subject to change."
           (function destroy-thread :replacement terminate-thread)))
 
 #!+sb-thread
-(defun enter-foreign-callback (arg1 arg2 arg3)
-  (initial-thread-function-trampoline
-   (make-foreign-thread :name "foreign callback")
-   nil #'sb!alien::enter-alien-callback nil t arg1 arg2 arg3))
+(defun enter-foreign-callback (index return arguments)
+  (let ((thread (without-gcing
+                  ;; Hold off GCing until *current-thread* is set up
+                  (setf *current-thread*
+                        (make-foreign-thread :name "foreign callback")))))
+    (dx-flet ((enter ()
+                (sb!alien::enter-alien-callback index return arguments)))
+      (initial-thread-function-trampoline thread nil #'enter nil))))
 
 (defmacro with-interruptions-lock ((thread) &body body)
   `(with-system-mutex ((thread-interruptions-lock ,thread))
@@ -1643,7 +1736,7 @@ subject to change."
       (funcall interruption))))
 
 #!+sb-thruption
-(defun run-interruption ()
+(defun run-interruption (*current-internal-error-context*)
   (in-interruption () ;the non-thruption code does this in the signal handler
     (let ((interruption (with-interruptions-lock (*current-thread*)
                           (pop (thread-interruptions *current-thread*)))))
@@ -1728,21 +1821,28 @@ Short version: be careful out there."
     (with-interrupts (funcall function)))
   #!-(and (not sb-thread) win32)
   (let ((os-thread (thread-os-thread thread)))
-    (cond ((not os-thread)
+    (cond ((= os-thread (ldb (byte sb!vm:n-word-bits 0) -1))
            (error 'interrupt-thread-error :thread thread))
           (t
-           (with-interruptions-lock (thread)
-             ;; Append to the end of the interruptions queue. It's
-             ;; O(N), but it does not hurt to slow interruptors down a
-             ;; bit when the queue gets long.
-             (setf (thread-interruptions thread)
-                   (append (thread-interruptions thread)
-                           (list (lambda ()
-                                   (without-interrupts
-                                     (allow-with-interrupts
-                                       (funcall function))))))))
-           (when (minusp (wake-thread os-thread))
-             (error 'interrupt-thread-error :thread thread))))))
+           (let (invoked)
+             (with-interruptions-lock (thread)
+               ;; Append to the end of the interruptions queue. It's
+               ;; O(N), but it does not hurt to slow interruptors down a
+               ;; bit when the queue gets long.
+               (setf (thread-interruptions thread)
+                     (append (thread-interruptions thread)
+                             (list (lambda ()
+                                     (setf invoked t)
+                                     (barrier (:memory))
+                                     (without-interrupts
+                                       (allow-with-interrupts
+                                         (funcall function))))))))
+             (when (and (minusp (wake-thread os-thread))
+                        ;; The interrupt queue has been processed by
+                        ;; some other interrupt.
+                        (progn (barrier (:memory))
+                               (not invoked)))
+               (error 'interrupt-thread-error :thread thread)))))))
 
 (defun terminate-thread (thread)
   "Terminate the thread identified by THREAD, by interrupting it and
@@ -1796,79 +1896,61 @@ assume that unknown code can safely be terminated using TERMINATE-THREAD."
 #!+sb-thread
 (progn
 
-  (sb!ext:defglobal sb!vm::*free-tls-index* 0)
+  (sb!ext:define-load-time-global sb!vm::*free-tls-index* 0)
   ;; Keep in sync with 'compiler/generic/parms.lisp'
   #!+ppc ; only PPC uses a separate symbol for the TLS index lock
-  (!defglobal sb!vm::*tls-index-lock* 0)
+  (!define-load-time-global sb!vm::*tls-index-lock* 0)
 
-  (defun %thread-sap (thread)
-    (let ((thread-sap (alien-sap (extern-alien "all_threads" (* t))))
-          (target (thread-os-thread thread)))
-      (loop
-        (when (sap= thread-sap (int-sap 0)) (return nil))
-        (let ((os-thread (sap-ref-word thread-sap
-                                       (* sb!vm:n-word-bytes
-                                          sb!vm::thread-os-thread-slot))))
-          (when (= os-thread target) (return thread-sap))
-          (setf thread-sap
-                (sap-ref-sap thread-sap (* sb!vm:n-word-bytes
-                                           sb!vm::thread-next-slot)))))))
-
-(defun %symbol-value-in-thread (symbol thread)
+  (defun %symbol-value-in-thread (symbol thread)
     ;; Prevent the thread from dying completely while we look for the TLS
     ;; area...
     (with-all-threads-lock
       (if (thread-alive-p thread)
-          (let* ((offset (get-lisp-obj-address (symbol-tls-index symbol)))
-                 (obj (sap-ref-lispobj (%thread-sap thread) offset))
-                 (tl-val (get-lisp-obj-address obj)))
-            (cond ((zerop offset)
-                   (values nil :no-tls-value))
-                  ((or (eql tl-val sb!vm:no-tls-value-marker-widetag)
-                       (eql tl-val sb!vm:unbound-marker-widetag))
-                   (values nil :unbound-in-thread))
-                  (t
-                   (values obj :ok))))
+          (let ((val (sap-ref-lispobj (int-sap (thread-primitive-thread thread))
+                                      (get-lisp-obj-address (symbol-tls-index symbol)))))
+            (case (get-lisp-obj-address val)
+              (#.sb!vm:no-tls-value-marker-widetag (values nil :no-tls-value))
+              (#.sb!vm:unbound-marker-widetag (values nil :unbound-in-thread))
+              (t (values val :ok))))
           (values nil :thread-dead))))
 
   (defun %set-symbol-value-in-thread (symbol thread value)
-    (with-pinned-objects (value)
-      ;; Prevent the thread from dying completely while we look for the TLS
-      ;; area...
-      (with-all-threads-lock
-        (if (thread-alive-p thread)
-            (let ((offset (get-lisp-obj-address (symbol-tls-index symbol))))
-              (cond ((zerop offset)
-                     (values nil :no-tls-value))
-                    (t
-                     (setf (sap-ref-lispobj (%thread-sap thread) offset)
-                           value)
-                     (values value :ok))))
-            (values nil :thread-dead)))))
-
-  (define-alien-variable tls-index-start unsigned-int)
+    ;; Prevent the thread from dying completely while we look for the TLS
+    ;; area...
+    (with-all-threads-lock
+      (if (thread-alive-p thread)
+          (let ((offset (get-lisp-obj-address (symbol-tls-index symbol))))
+            (cond ((zerop offset)
+                   (values nil :no-tls-value))
+                  (t
+                   (setf (sap-ref-lispobj (int-sap (thread-primitive-thread thread))
+                                          offset)
+                         value)
+                   (values value :ok))))
+          (values nil :thread-dead))))
 
   ;; Get values from the TLS area of the current thread.
+  ;; Disregard duplicates and immediate objects.
   (defun %thread-local-references ()
     ;; TLS-INDEX-START is a word number relative to thread base.
     ;; *FREE-TLS-INDEX* - which is only manipulated by machine code  - is an
-    ;; offset from thread base to the next usable TLS cell as a raw word
-    ;; manifesting in Lisp as a fixnum. Convert it from byte number to word
-    ;; number before using it as the upper bound for the sequential scan.
-    (loop for index from tls-index-start
-          ;; On x86-64, mask off the sign bit. It is used as a semaphore.
-          below (ash (logand #!+x86-64 most-positive-fixnum
-                             sb!vm::*free-tls-index*)
-                     (+ (- sb!vm:word-shift) sb!vm:n-fixnum-tag-bits))
-          for obj = (sb!kernel:%make-lisp-obj
-                     (sb!sys:sap-int (sb!vm::current-thread-offset-sap index)))
-          unless (or (member (widetag-of obj)
-                             `(,sb!vm:no-tls-value-marker-widetag
-                               ,sb!vm:unbound-marker-widetag))
-                     (typep obj '(or fixnum character))
-                     (memq obj seen))
-          collect obj into seen
-          finally (return seen))))
+    ;; offset from thread base to the next usable TLS cell as a byte offset
+    ;; (raw value) manifesting in Lisp as a fixnum.
+    ;; The sign bit of sb!vm::*free-tls-index* is a semaphore,
+    ;; except on PPC where it isn't, but masking is fine in any case.
+    (do ((index (- (ash (logand sb!vm::*free-tls-index* most-positive-fixnum)
+                        sb!vm:n-fixnum-tag-bits)
+                   sb!vm:n-word-bytes)
+                (- index sb!vm:n-word-bytes))
+         ;; (There's no reason this couldn't work on any thread now.)
+         (sap (int-sap (thread-primitive-thread *current-thread*)))
+         (list))
+        ((< index (ash tls-index-start sb!vm:word-shift)) list)
+      (let ((obj (sap-ref-lispobj sap index)))
+        (when (and obj ; don't bother returning NIL
+                   (sb!vm:is-lisp-pointer (get-lisp-obj-address obj))
+                   (not (memq obj list)))
+          (push obj list))))))
 
 (defun symbol-value-in-thread (symbol thread &optional (errorp t))
   "Return the local value of SYMBOL in THREAD, and a secondary value of T
