@@ -11,8 +11,8 @@
 
 (in-package "SB!VM")
 
-(defconstant arg-count-sc (make-sc-offset immediate-arg-scn nargs-offset))
-(defconstant closure-sc (make-sc-offset descriptor-reg-sc-number lexenv-offset))
+(defconstant arg-count-sc (make-sc+offset immediate-arg-scn nargs-offset))
+(defconstant closure-sc (make-sc+offset descriptor-reg-sc-number lexenv-offset))
 
 ;;; Make a passing location TN for a local call return PC. If standard
 ;;; is true, then use the standard (full call) location, otherwise use
@@ -28,10 +28,11 @@
 ;;; standard convention, but is totally unrestricted in non-standard
 ;;; conventions, since we can always fetch it off of the stack using
 ;;; the arg pointer.
-(defun make-old-fp-passing-location (standard)
-  (if standard
-      (make-wired-tn *fixnum-primitive-type* immediate-arg-scn ocfp-offset)
-      (make-normal-tn *fixnum-primitive-type*)))
+(defun make-old-fp-passing-location ()
+  (make-wired-tn *fixnum-primitive-type* immediate-arg-scn ocfp-offset))
+
+(defconstant old-fp-passing-offset
+  (make-sc+offset descriptor-reg-sc-number ocfp-offset))
 
 ;;; These functions make the TNs used to hold Old-FP and Return-PC
 ;;; within the current function. We treat these specially so that the
@@ -573,10 +574,12 @@ default-value-8
      (:args
       ,@(unless (eq return :tail)
           '((new-fp :scs (any-reg) :to :eval)))
-
-      ,(if named
-           '(name :target name-pass)
-           '(arg-fun :target lexenv))
+      ,@(case named
+          ((nil)
+           '((arg-fun :target lexenv)))
+          (:direct)
+          (t
+           '((name :target name-pass))))
 
       ,@(when (eq return :tail)
           '((ocfp :target ocfp-pass)
@@ -595,6 +598,7 @@ default-value-8
      (:vop-var vop)
      (:info ,@(unless (or variable (eq return :tail)) '(arg-locs))
             ,@(unless variable '(nargs))
+            ,@(when (eq named :direct) '(fun))
             ,@(when (eq return :fixed) '(nvals))
             step-instrumenting)
 
@@ -618,19 +622,20 @@ default-value-8
                   :to :eval)
                  return-pc-pass)
 
-     ,@(if named
-         `((:temporary (:sc descriptor-reg :offset fdefn-offset
-                        :from (:argument ,(if (eq return :tail) 0 1))
-                        :to :eval)
-                       name-pass))
-
-         `((:temporary (:sc descriptor-reg :offset lexenv-offset
-                        :from (:argument ,(if (eq return :tail) 0 1))
-                        :to :eval)
-                       lexenv)
-           #!-gengc
-           (:temporary (:scs (descriptor-reg) :from (:argument 0) :to :eval)
-                       function)))
+     ,@(case named
+         ((t)
+          `((:temporary (:sc descriptor-reg :offset fdefn-offset
+                         :from (:argument ,(if (eq return :tail) 0 1))
+                         :to :eval)
+                        name-pass)))
+         ((nil)
+          `((:temporary (:sc descriptor-reg :offset lexenv-offset
+                         :from (:argument ,(if (eq return :tail) 0 1))
+                         :to :eval)
+                        lexenv)
+            #!-gengc
+            (:temporary (:scs (descriptor-reg) :from (:argument 0) :to :eval)
+                        function))))
 
      (:temporary (:sc any-reg :offset nargs-offset :to :eval)
                  nargs-pass)
@@ -729,50 +734,53 @@ default-value-8
                                          (move new-fp cfp-tn)
                                          (move csp-tn cfp-tn))))))
                       ((nil))))))
-
-           ,@(if named
-                 `((sc-case name
-                     (descriptor-reg (move name name-pass))
-                     (control-stack
-                      (inst ldl name-pass
-                            (ash (tn-offset name) word-shift) cfp-tn)
-                      (do-next-filler))
-                     (constant
-                      (inst ldl name-pass
-                            (- (ash (tn-offset name) word-shift)
-                               other-pointer-lowtag) code-tn)
-                      (do-next-filler)))
-                   (inst ldl entry-point
-                         (- (ash fdefn-raw-addr-slot word-shift)
-                            other-pointer-lowtag) name-pass)
-                   (do-next-filler))
-                 `((sc-case arg-fun
-                     (descriptor-reg (move arg-fun lexenv))
-                     (control-stack
-                      (inst ldl lexenv
-                            (ash (tn-offset arg-fun) word-shift) cfp-tn)
-                      (do-next-filler))
-                     (constant
-                      (inst ldl lexenv
-                            (- (ash (tn-offset arg-fun) word-shift)
-                               other-pointer-lowtag) code-tn)
-                      (do-next-filler)))
-                   #!-gengc
-                   (inst ldl function
-                         (- (ash closure-fun-slot word-shift)
-                            fun-pointer-lowtag) lexenv)
-                   #!-gengc
-                   (do-next-filler)
-                   #!-gengc
-                   (inst addq function
-                         (- (ash simple-fun-code-offset word-shift)
-                            fun-pointer-lowtag) entry-point)
-                   #!+gengc
-                   (inst ldl entry-point
-                         (- (ash closure-entry-point-slot word-shift)
-                            fun-pointer-lowtag) lexenv)
-                   #!+gengc
-                   (do-next-filler)))
+           ,@(case named
+               ((t)
+                `((sc-case name
+                    (descriptor-reg (move name name-pass))
+                    (control-stack
+                     (inst ldl name-pass
+                           (ash (tn-offset name) word-shift) cfp-tn)
+                     (do-next-filler))
+                    (constant
+                     (inst ldl name-pass
+                           (- (ash (tn-offset name) word-shift)
+                              other-pointer-lowtag) code-tn)
+                     (do-next-filler)))
+                  (inst ldl entry-point
+                        (- (ash fdefn-raw-addr-slot word-shift)
+                           other-pointer-lowtag) name-pass)
+                  (do-next-filler)))
+               ((nil)
+                `((sc-case arg-fun
+                    (descriptor-reg (move arg-fun lexenv))
+                    (control-stack
+                     (inst ldl lexenv
+                           (ash (tn-offset arg-fun) word-shift) cfp-tn)
+                     (do-next-filler))
+                    (constant
+                     (inst ldl lexenv
+                           (- (ash (tn-offset arg-fun) word-shift)
+                              other-pointer-lowtag) code-tn)
+                     (do-next-filler)))
+                  #!-gengc
+                  (inst ldl function
+                        (- (ash closure-fun-slot word-shift)
+                           fun-pointer-lowtag) lexenv)
+                  #!-gengc
+                  (do-next-filler)
+                  #!-gengc
+                  (inst addq function
+                        (- (ash simple-fun-code-offset word-shift)
+                           fun-pointer-lowtag) entry-point)
+                  #!+gengc
+                  (inst ldl entry-point
+                        (- (ash closure-entry-point-slot word-shift)
+                           fun-pointer-lowtag) lexenv)
+                  #!+gengc
+                  (do-next-filler)))
+               (:direct
+                `((inst ldl entry-point (static-fun-offset fun) null-tn))))
            (loop
              (if (cdr filler)
                  (do-next-filler)
@@ -798,10 +806,13 @@ default-value-8
 
 (define-full-call call nil :fixed nil)
 (define-full-call call-named t :fixed nil)
+(define-full-call static-call-named :direct :fixed nil)
 (define-full-call multiple-call nil :unknown nil)
 (define-full-call multiple-call-named t :unknown nil)
+(define-full-call static-multiple-call-named :direct :unknown nil)
 (define-full-call tail-call nil :tail nil)
 (define-full-call tail-call-named t :tail nil)
+(define-full-call static-tail-call-named :direct :tail nil)
 
 (define-full-call call-variable nil :fixed t)
 (define-full-call multiple-call-variable nil :unknown t)
@@ -1144,33 +1155,12 @@ default-value-8
     (inst subq supplied (fixnumize fixed) count)
     (inst subq csp-tn count context)))
 
-;;; Signal wrong argument count error if NARGS isn't equal to COUNT.
-#!-precise-arg-count-error
-(define-vop (verify-arg-count)
-  (:policy :fast-safe)
-  (:translate sb!c::%verify-arg-count)
-  (:args (nargs :scs (any-reg)))
-  (:arg-types positive-fixnum (:constant t))
-  (:temporary (:scs (any-reg) :type fixnum) temp)
-  (:info count)
-  (:vop-var vop)
-  (:save-p :compute-only)
-  (:generator 3
-    (let ((err-lab
-           (generate-error-code vop 'invalid-arg-count-error nargs)))
-      (cond ((zerop count)
-             (inst bne nargs err-lab))
-            (t
-             (inst subq nargs (fixnumize count) temp)
-             (inst bne temp err-lab))))))
-
-#!+precise-arg-count-error
 (define-vop (verify-arg-count)
   (:policy :fast-safe)
   (:args (nargs :scs (any-reg)))
   (:arg-types positive-fixnum (:constant t) (:constant t))
   (:info min max)
-  (:temporary (:scs (any-reg) :type fixnum) temp)
+  (:temporary (:scs (unsigned-reg) :type fixnum) temp)
   (:vop-var vop)
   (:save-p :compute-only)
   (:generator 3
@@ -1180,7 +1170,7 @@ default-value-8
              (cond ((zerop max)
                     (inst bne nargs err-lab))
                    (t
-                    (inst subq nargs (fixnumize count) temp)
+                    (inst subq nargs (fixnumize max) temp)
                     (inst bne temp err-lab))))
             (max
              (when (plusp min)
