@@ -45,11 +45,13 @@
            (- list-pointer-lowtag)))
       0))
 
+(defconstant-eqx +all-static-fdefns+
+    #.(concatenate 'vector +c-callable-fdefns+ +static-fdefns+) #'equalp)
+
 ;;; Return the (byte) offset from NIL to the start of the fdefn object
 ;;; for the static function NAME.
 (defun static-fdefn-offset (name)
-  (let ((static-fun-index
-          (position name #.(concatenate 'vector +c-callable-fdefns+ +static-fdefns+))))
+  (let ((static-fun-index (position name +all-static-fdefns+)))
     (and static-fun-index
          (+ (* (length +static-symbols+) (pad-data-block symbol-size))
             (pad-data-block (1- symbol-size))
@@ -72,7 +74,7 @@
      (* fdefn-raw-addr-slot n-word-bytes)))
 
 ;;; Various error-code generating helpers
-(defvar *adjustable-vectors* nil)
+(defvar *adjustable-vectors*)
 
 (defmacro with-adjustable-vector ((var) &rest body)
   `(let ((,var (or (pop *adjustable-vectors*)
@@ -86,10 +88,13 @@
      ;; more than 16 elements. Maybe we should make it nonadjustable?
      (declare (type (vector (unsigned-byte 8)) ,var))
      (setf (fill-pointer ,var) 0)
-     (unwind-protect
-         (progn
-           ,@body)
-       (push ,var *adjustable-vectors*))))
+     ;; No UNWIND-PROTECT here - semantics are unaffected by nonlocal exit,
+     ;; and this macro is about speeding up the compiler, not slowing it down.
+     ;; GC will clean up any debris, and since the vector does not point
+     ;; to anything, even an accidental promotion to a higher generation
+     ;; will not cause transitive garbage retention.
+     ,@body
+     (push ,var *adjustable-vectors*)))
 
 ;;;; interfaces to IR2 conversion
 
@@ -102,12 +107,23 @@
                      (nth n *register-arg-offsets*))
       (make-wired-tn *backend-t-primitive-type* control-stack-sc-number n)))
 
+;;; Same as above but marks stack locations as :arg-pass
+(defun standard-call-arg-location (n)
+  (declare (type unsigned-byte n))
+  (if (< n register-arg-count)
+      (make-wired-tn *backend-t-primitive-type* descriptor-reg-sc-number
+                     (nth n *register-arg-offsets*))
+      (let ((tn
+              (make-wired-tn *backend-t-primitive-type* control-stack-sc-number n)))
+        (setf (tn-kind tn) :arg-pass)
+        tn)))
+
 (defun standard-arg-location-sc (n)
   (declare (type unsigned-byte n))
   (if (< n register-arg-count)
-      (make-sc-offset descriptor-reg-sc-number
+      (make-sc+offset descriptor-reg-sc-number
                       (nth n *register-arg-offsets*))
-      (make-sc-offset control-stack-sc-number n)))
+      (make-sc+offset control-stack-sc-number n)))
 
 ;;; Make a TN to hold the number-stack frame pointer.  This is allocated
 ;;; once per component, and is component-live.
@@ -124,7 +140,9 @@
    (make-representation-tn *fixnum-primitive-type* any-reg-sc-number)
    env))
 
-(defun make-stack-pointer-tn ()
+#!-x86-64
+(defun make-stack-pointer-tn (&optional nargs)
+  (declare (ignore nargs))
   (make-normal-tn *fixnum-primitive-type*))
 
 (defun make-number-stack-pointer-tn ()
@@ -135,9 +153,14 @@
 
 ;;; Return a list of TNs that can be used to represent an unknown-values
 ;;; continuation within a function.
-(defun make-unknown-values-locations ()
+(defun make-unknown-values-locations (&optional unused-count)
+  (declare (ignorable unused-count))
   (list (make-stack-pointer-tn)
-        (make-normal-tn *fixnum-primitive-type*)))
+        (cond #!+x86-64 ;; needs support from receive-unknown-values
+              (unused-count
+               (sb!c::make-unused-tn))
+              (t
+               (make-normal-tn *fixnum-primitive-type*)))))
 
 ;;; This function is called by the ENTRY-ANALYZE phase, allowing
 ;;; VM-dependent initialization of the IR2-COMPONENT structure. We
@@ -150,4 +173,17 @@
                         (ir2-component-constants (component-info component))))
   (values))
 
+(defun error-call (vop error-code &rest values)
+  "Cause an error.  ERROR-CODE is the error to cause."
+  (emit-error-break vop error-trap (error-number-or-lose error-code) values))
 
+(defun cerror-call (vop error-code &rest values)
+  "Cause a continuable error.  ERROR-CODE is the error to cause."
+  (emit-error-break vop cerror-trap (error-number-or-lose error-code) values))
+
+#!+sb-safepoint
+(define-vop (insert-safepoint)
+  (:policy :fast-safe)
+  (:translate sb!kernel::gc-safepoint)
+  (:generator 0
+    (emit-safepoint)))
