@@ -11,8 +11,8 @@
 
 (in-package "SB!VM")
 
-(defconstant arg-count-sc (make-sc-offset immediate-arg-scn nargs-offset))
-(defconstant closure-sc (make-sc-offset descriptor-reg-sc-number lexenv-offset))
+(defconstant arg-count-sc (make-sc+offset immediate-arg-scn nargs-offset))
+(defconstant closure-sc (make-sc+offset descriptor-reg-sc-number lexenv-offset))
 
 ;;; Make a passing location TN for a local call return PC.  If standard is
 ;;; true, then use the standard (full call) location, otherwise use any legal
@@ -28,10 +28,11 @@
 ;;; standard convention, but is totally unrestricted in non-standard
 ;;; conventions, since we can always fetch it off of the stack using
 ;;; the arg pointer.
-(defun make-old-fp-passing-location (standard)
-  (if standard
-      (make-wired-tn *fixnum-primitive-type* immediate-arg-scn ocfp-offset)
-      (make-normal-tn *fixnum-primitive-type*)))
+(defun make-old-fp-passing-location ()
+  (make-wired-tn *fixnum-primitive-type* immediate-arg-scn ocfp-offset))
+
+(defconstant old-fp-passing-offset
+  (make-sc+offset descriptor-reg-sc-number ocfp-offset))
 
 ;;; Make the TNs used to hold OLD-FP and RETURN-PC within the current
 ;;; function. We treat these specially so that the debugger can find
@@ -767,12 +768,8 @@ default-value-8
                     ;; CONTEXT-PC will be pointing here when the
                     ;; interrupt is handled, not after the BREAK.
                     (note-this-location vop :step-before-vop)
-                    ;; Construct a trap code with the low bits from
-                    ;; SINGLE-STEP-AROUND-TRAP and the high bits from
-                    ;; the register number of CALLABLE-TN.
-                    (inst break 0 (logior single-step-around-trap
-                                          (ash (reg-tn-encoding callable-tn)
-                                               5)))
+                    (inst break (reg-tn-encoding callable-tn)
+                          single-step-around-trap)
                     (emit-label step-done-label))))
            (declare (ignorable #'insert-step-instrumenting))
            ,@(case named
@@ -1191,64 +1188,39 @@ default-value-8
     (inst addu count supplied (fixnumize (- fixed)))
     (inst subu context csp-tn count)))
 
-
-;;; Signal wrong argument count error if Nargs isn't = to Count.
-;;;
-#!-precise-arg-count-error
-(define-vop (verify-arg-count)
-  (:policy :fast-safe)
-  (:translate sb!c::%verify-arg-count)
-  (:args (nargs :scs (any-reg)))
-  (:arg-types positive-fixnum (:constant t))
-  (:temporary (:scs (any-reg) :type fixnum) temp)
-  (:info count)
-  (:vop-var vop)
-  (:save-p :compute-only)
-  (:generator 3
-    (let ((err-lab
-           (generate-error-code vop 'invalid-arg-count-error nargs)))
-      (cond ((zerop count)
-             (inst bne nargs err-lab)
-             (inst nop))
-            (t
-             (inst li temp (fixnumize count))
-             (inst bne nargs temp err-lab)
-             (inst nop))))))
-
-#!+precise-arg-count-error
 (define-vop (verify-arg-count)
   (:policy :fast-safe)
   (:args (nargs :scs (any-reg)))
-  (:temporary (:scs (any-reg) :type fixnum) temp)
+  (:temporary (:scs (unsigned-reg)) temp)
   (:arg-types positive-fixnum (:constant t) (:constant t))
   (:info min max)
   (:vop-var vop)
   (:save-p :compute-only)
   (:generator 3
-              (let ((err-lab
-                      (generate-error-code vop 'invalid-arg-count-error nargs)))
-                (cond ((not min)
-                       (cond ((zerop max)
-                              (inst bne nargs err-lab))
-                             (t
-                              (inst li temp (fixnumize max))
-                              (inst bne nargs temp err-lab)))
-                       (inst nop))
-                      (max
-                       (when (plusp min)
-                         (inst li temp (fixnumize min))
-                         (inst sltu temp nargs temp)
-                         (inst bne temp err-lab)
-                         (inst nop))
-                       (inst li temp (fixnumize max))
-                       (inst sltu temp temp nargs)
-                       (inst bne temp err-lab)
-                       (inst nop))
-                      ((plusp min)
-                       (inst li temp (fixnumize min))
-                       (inst sltu temp nargs temp)
-                       (inst bne temp err-lab)
-                       (inst nop))))))
+    (let ((err-lab
+            (generate-error-code vop 'invalid-arg-count-error nargs)))
+      (cond ((not min)
+             (cond ((zerop max)
+                    (inst bne nargs err-lab))
+                   (t
+                    (inst li temp (fixnumize max))
+                    (inst bne nargs temp err-lab)))
+             (inst nop))
+            (max
+             (when (plusp min)
+               (inst li temp (fixnumize min))
+               (inst sltu temp nargs temp)
+               (inst bne temp err-lab)
+               (inst nop))
+             (inst li temp (fixnumize max))
+             (inst sltu temp temp nargs)
+             (inst bne temp err-lab)
+             (inst nop))
+            ((plusp min)
+             (inst li temp (fixnumize min))
+             (inst sltu temp nargs temp)
+             (inst bne temp err-lab)
+             (inst nop))))))
 
 ;;; Single-stepping
 
