@@ -11,6 +11,162 @@
 
 (in-package "SB!VM")
 
+;;;; allocation helpers
+
+;;; Most allocation is done by inline code with sometimes help
+;;; from the C alloc() function by way of the alloc-tramp
+;;; assembly routine.
+
+(defun allocation-dynamic-extent (alloc-tn size lowtag)
+  (aver (not (location= alloc-tn rsp-tn)))
+  (inst sub rsp-tn size)
+  ;; see comment in x86/macros.lisp implementation of this
+  ;; However that comment seems inapplicable here because:
+  ;; - PAD-DATA-BLOCK quite clearly enforces double-word alignment,
+  ;;   contradicting "... unfortunately not enforced by ..."
+  ;; - It's not the job of FIXED-ALLOC to realign anything.
+  ;; - The real issue is that it's not obvious that the stack is
+  ;;   16-byte-aligned at *all* times. Maybe it is, maybe it isn't.
+  (inst and rsp-tn #.(lognot lowtag-mask))
+  (inst lea alloc-tn (make-ea :byte :base rsp-tn :disp lowtag))
+  (values))
+
+(defun allocation-tramp (node result-tn size lowtag)
+  (cond ((typep size '(and integer (not (signed-byte 32))))
+         ;; MOV accepts large immediate operands, PUSH does not
+         (inst mov result-tn size)
+         (inst push result-tn))
+        (t
+         (inst push size)))
+  ;; This really would be better if it recognized TEMP-REG-TN as the "good" case
+  ;; rather than specializing on R11, which just happens to be the temp reg.
+  ;; But the assembly routine is hand-written, not generated, and it has to match,
+  ;; so there's not much that can be done to generalize it.
+  (let ((to-r11 (location= result-tn r11-tn)))
+    (invoke-asm-routine 'call (if to-r11 'alloc-tramp-r11 'alloc-tramp) node)
+    (unless to-r11
+      (inst pop result-tn)))
+
+  (when lowtag
+    (inst or (reg-in-size result-tn :byte) lowtag))
+  (values))
+
+;;; Emit code to allocate an object with a size in bytes given by
+;;; SIZE into ALLOC-TN. The size may be an integer of a TN.
+;;; NODE may be used to make policy-based decisions.
+;;; This function should only be used inside a pseudo-atomic section,
+;;; which to the degree needed should also cover subsequent initialization.
+(defun allocation (alloc-tn size node &optional dynamic-extent lowtag)
+  (when dynamic-extent
+    (allocation-dynamic-extent alloc-tn size lowtag)
+    (return-from allocation (values)))
+  (aver (and (not (location= alloc-tn temp-reg-tn))
+             (or (integerp size) (not (location= size temp-reg-tn)))))
+
+  #!+(and (not sb-thread) sb-dynamic-core)
+  ;; We'd need a spare reg in which to load boxed_region from the linkage table.
+  ;; Could push/pop any random register on the stack and own it temporarily,
+  ;; but seeing as nobody cared about this, just punt.
+  (allocation-tramp node alloc-tn size lowtag)
+
+  #!-(and (not sb-thread) sb-dynamic-core)
+  ;; Otherwise do the normal inline allocation thing
+  (let ((NOT-INLINE (gen-label))
+        (DONE (gen-label))
+        (SKIP-INSTRUMENTATION (gen-label))
+        ;; Yuck.
+        (in-elsewhere (eq *elsewhere* sb!assem::**current-segment**))
+        ;; thread->alloc_region.free_pointer
+        (free-pointer
+         #!+sb-thread
+         (thread-tls-ea (* n-word-bytes thread-alloc-region-slot))
+         #!-sb-thread
+         (make-ea :qword :disp (make-fixup "gc_alloc_region" :foreign)))
+        ;; thread->alloc_region.end_addr
+        (end-addr
+         #!+sb-thread
+         (thread-tls-ea (* n-word-bytes (1+ thread-alloc-region-slot)))
+         #!-sb-thread
+         (make-ea :qword :disp (make-fixup "gc_alloc_region" :foreign 8))))
+
+    ;; Insert allocation profiler instrumentation
+    ;; FIXME: for now, change this to '>=' to perform self-build
+    ;;        where the resulting executable has instrumentation
+    ;;        because I can't get policies to work.
+    ;; FIXME: and does this work for assembly routines?
+    (when (policy node (> sb!c::instrument-consing 1))
+      (inst mov temp-reg-tn
+            (make-ea :qword :base thread-base-tn
+                     :disp (* n-word-bytes thread-profile-data-slot)))
+      (inst test temp-reg-tn temp-reg-tn)
+      ;; This instruction is modified to "JMP :z" when profiling is
+      ;; partially enabled. After the buffer is assigned, it becomes
+      ;; fully enabled. The unconditional jmp gives minimal performance
+      ;; loss if the profiler is statically disabled. (one memory
+      ;; read and a test whose result is never used, which the CPU
+      ;; is good at ignoring as far as instruction prefetch goes)
+      (inst jmp skip-instrumentation)
+      (emit-alignment 3 :long-nop)
+      (let ((helper (if (integerp size)
+                        'enable-alloc-counter
+                        'enable-sized-alloc-counter)))
+        (cond ((or (not node) ; assembly routine
+                   (sb!c::code-immobile-p node))
+               (inst call (make-fixup helper :assembly-routine)) ; 5 bytes
+               (emit-long-nop sb!assem::**current-segment** 3)) ; align
+              (t
+               (inst call ; 7 bytes
+                     (make-ea :qword :disp
+                              (make-fixup helper :assembly-routine*)))
+               (inst nop))) ; align
+        (unless (integerp size)
+          ;; This TEST instruction is never executed- it informs the profiler
+          ;; which register holds SIZE.
+          (inst test size size) ; 3 bytes
+          (emit-long-nop sb!assem::**current-segment** 5))) ; align
+      (emit-label skip-instrumentation))
+
+    (cond ((or in-elsewhere
+               ;; large objects will never be made in a per-thread region
+               (and (integerp size)
+                    (>= size large-object-size)))
+           (allocation-tramp node alloc-tn size lowtag))
+          (t
+           (inst mov temp-reg-tn free-pointer)
+           (cond ((integerp size)
+                  (inst lea alloc-tn (make-ea :qword :base temp-reg-tn :disp size)))
+                 ((location= alloc-tn size)
+                  (inst add alloc-tn temp-reg-tn))
+                 (t
+                  (inst lea alloc-tn (make-ea :qword :base temp-reg-tn :index size))))
+           (inst cmp alloc-tn end-addr)
+           (inst jmp :a NOT-INLINE)
+           (inst mov free-pointer alloc-tn)
+           (emit-label DONE)
+           (if lowtag
+               (inst lea alloc-tn (make-ea :byte :base temp-reg-tn :disp lowtag))
+               (inst mov alloc-tn temp-reg-tn))
+           (assemble (*elsewhere*)
+             (emit-label NOT-INLINE)
+             (cond ((and (tn-p size) (location= size alloc-tn)) ; recover SIZE
+                    (inst sub alloc-tn free-pointer)
+                    (allocation-tramp node temp-reg-tn alloc-tn nil))
+                   (t ; SIZE is intact
+                    (allocation-tramp node temp-reg-tn size nil)))
+             (inst jmp DONE))))
+    (values)))
+
+;;; Allocate an other-pointer object of fixed SIZE with a single word
+;;; header having the specified WIDETAG value. The result is placed in
+;;; RESULT-TN.
+(defun fixed-alloc (result-tn widetag size node &optional stack-allocate-p)
+  (maybe-pseudo-atomic stack-allocate-p
+      (allocation result-tn (pad-data-block size) node stack-allocate-p
+                  other-pointer-lowtag)
+      (storew* (logior (ash (1- size) n-widetag-bits) widetag)
+               result-tn 0 other-pointer-lowtag
+               (not stack-allocate-p))))
+
 ;;;; CONS, LIST and LIST*
 (define-vop (list-or-list*)
   (:args (things :more t))
@@ -118,38 +274,71 @@
     (:args (type :scs (unsigned-reg immediate))
            (length :scs (any-reg immediate))
            (words :scs (any-reg immediate)))
+    ;; Result is live from the beginning, like a temp, because we use it as such
+    ;; in 'calc-size-in-bytes'
     (:results (result :scs (descriptor-reg) :from :load))
     (:arg-types positive-fixnum positive-fixnum positive-fixnum)
     (:policy :fast-safe)
+    (:node-var node)
     (:generator 100
       ;; The LET generates instructions that needn't be pseudoatomic
       ;; so don't move it inside.
       (let ((size (calc-size-in-bytes words result)))
         (pseudo-atomic
-         (allocation result size nil nil other-pointer-lowtag)
+         (allocation result size node nil other-pointer-lowtag)
          (put-header result type length t)))))
 
   (define-vop (allocate-vector-on-stack)
-    (:args (type :scs (unsigned-reg immediate) :to :save)
-           (length :scs (any-reg immediate) :to :eval :target rax)
-           (words :scs (any-reg immediate) :target rcx))
-    (:temporary (:sc any-reg :offset ecx-offset :from (:argument 2)) rcx)
-    (:temporary (:sc any-reg :offset eax-offset :from :eval) rax)
-    (:temporary (:sc any-reg :offset edi-offset) rdi)
-    (:temporary (:sc complex-double-reg) zero)
+    (:args (type :scs (unsigned-reg immediate))
+           (length :scs (any-reg immediate))
+           (words :scs (any-reg immediate)))
     (:results (result :scs (descriptor-reg) :from :load))
+    (:temporary (:sc any-reg :offset ecx-offset :from :eval) rcx)
+    (:temporary (:sc any-reg :offset eax-offset :from :eval) rax)
+    (:temporary (:sc any-reg :offset edi-offset :from :eval) rdi)
+    (:temporary (:sc complex-double-reg) zero)
     (:arg-types positive-fixnum positive-fixnum positive-fixnum)
     (:translate allocate-vector)
     (:policy :fast-safe)
     (:node-var node)
     (:generator 100
-      (let ((size (calc-size-in-bytes words result)))
+      (let ((size (calc-size-in-bytes words result))
+            (rax-zeroed))
         (allocation result size node t other-pointer-lowtag)
         (put-header result type length nil)
         ;; FIXME: It would be good to check for stack overflow here.
         ;; It would also be good to skip zero-fill of specialized vectors
         ;; perhaps in a policy-dependent way. At worst you'd see random
         ;; bits, and CLHS says consequences are undefined.
+        (when sb!c::*msan-unpoison*
+          ;; Unpoison all DX vectors regardless of widetag.
+          ;; Mark the header and length as valid, not just the payload.
+          #!+linux ; unimplemented for others
+          (let ((words-savep
+                 ;; 'words' might be co-located with any of the temps
+                 (or (location= words rdi) (location= words rcx) (location= words rax)))
+                (rax rax))
+            (setq rax-zeroed (not (location= words rax)))
+            (when words-savep ; use 'result' to save 'words'
+              (inst mov result words))
+            (cond ((sc-is words immediate)
+                   (inst mov rcx (+ (tn-value words) vector-data-offset)))
+                  (t
+                   (inst lea rcx
+                         (make-ea :qword :base words
+                                  :disp (ash vector-data-offset n-fixnum-tag-bits)))
+                   (if (= n-fixnum-tag-bits 1)
+                       (setq rax (reg-in-size rax :dword)) ; don't bother shifting rcx
+                       (inst shr rcx n-fixnum-tag-bits))))
+            (inst mov rdi msan-mem-to-shadow-xor-const)
+            (inst xor rdi rsp-tn) ; compute shadow address
+            (zeroize rax)
+            (inst rep)
+            (inst stos rax)
+            (when words-savep
+              (inst mov words result) ; restore 'words'
+              (inst lea result ; recompute the tagged pointer
+                    (make-ea :byte :base rsp-tn :disp other-pointer-lowtag)))))
         (let ((data-addr
                 (make-ea :qword :base result
                                 :disp (- (* vector-data-offset n-word-bytes)
@@ -160,8 +349,7 @@
                      (cond ((> n 8)
                             (inst mov rcx (tn-value words)))
                            ((= n 1)
-                            (zeroize rax)
-                            (inst mov data-addr rax)
+                            (inst mov data-addr 0)
                             (return-from zero-fill))
                            (t
                             (multiple-value-bind (double single) (truncate n 2)
@@ -177,7 +365,7 @@
                    (move rcx words)
                    (inst shr rcx n-fixnum-tag-bits)))
             (inst lea rdi data-addr)
-            (zeroize rax)
+            (unless rax-zeroed (zeroize rax))
             (inst rep)
             (inst stos rax)))))))
 
@@ -274,11 +462,11 @@
   (:results (result :scs (descriptor-reg) :from :argument))
   (:node-var node)
   (:generator 37
-    (with-fixed-allocation (result fdefn-widetag fdefn-size node)
-      (storew name result fdefn-name-slot other-pointer-lowtag)
-      (storew nil-value result fdefn-fun-slot other-pointer-lowtag)
-      (storew (make-fixup 'undefined-tramp :assembly-routine)
-              result fdefn-raw-addr-slot other-pointer-lowtag))))
+    (fixed-alloc result fdefn-widetag fdefn-size node)
+    (storew name result fdefn-name-slot other-pointer-lowtag)
+    (storew nil-value result fdefn-fun-slot other-pointer-lowtag)
+    (storew (make-fixup 'undefined-tramp :assembly-routine)
+            result fdefn-raw-addr-slot other-pointer-lowtag)))
 
 (define-vop (make-closure)
   ; (:args (function :to :save :scs (descriptor-reg)))
@@ -288,22 +476,30 @@
   (:node-var node)
   (:generator 10
    (maybe-pseudo-atomic stack-allocate-p
-    (let* ((size (+ length closure-info-offset))
-           (header (logior (ash (1- size) n-widetag-bits) closure-widetag)))
-      (allocation result (pad-data-block size) node stack-allocate-p
-                  fun-pointer-lowtag)
-      (storew* #!-immobile-space header ; write the widetag and size
-               #!+immobile-space        ; ... plus the layout pointer
-               (progn (inst mov temp header)
-                      (inst or temp (static-symbol-value-ea 'function-layout))
-                      temp)
-               result 0 fun-pointer-lowtag (not stack-allocate-p)))
-    ;; These two instructions are within the scope of PSEUDO-ATOMIC.
-    ;; This is due to scav_closure() assuming that it can always subtract
-    ;; FUN_RAW_ADDR_OFFSET from closure->fun to obtain a Lisp object,
-    ;; without any precheck for whether that word is currently 0.
-    (inst lea temp (make-fixup nil :closure label))
-    (storew temp result closure-fun-slot fun-pointer-lowtag))))
+     (let* ((size (+ length closure-info-offset))
+            (header (logior (ash (1- size) n-widetag-bits) closure-widetag)))
+       (allocation result (pad-data-block size) node stack-allocate-p
+                   fun-pointer-lowtag)
+       (storew* #!-immobile-space header ; write the widetag and size
+                #!+immobile-space        ; ... plus the layout pointer
+                (progn (inst mov temp header)
+                       (inst or temp #!-sb-thread (static-symbol-value-ea 'function-layout)
+                                     #!+sb-thread
+                                     (thread-tls-ea (ash thread-function-layout-slot
+                                                         word-shift)))
+                       temp)
+                result 0 fun-pointer-lowtag (not stack-allocate-p)))
+     ;; These two instructions are within the scope of PSEUDO-ATOMIC.
+     ;; This is due to scav_closure() assuming that it can always subtract
+     ;; FUN_RAW_ADDR_OFFSET from closure->fun to obtain a Lisp object,
+     ;; without any precheck for whether that word is currently 0.
+     (inst lea (reg-in-size temp :immobile-code-pc) (make-fixup nil :closure label))
+     (storew (reg-in-size temp
+                          (if stack-allocate-p
+                              ;; Need to do a full word store because the stack not zeroed
+                              :qword
+                              :immobile-code-pc))
+             result closure-fun-slot fun-pointer-lowtag))))
 
 ;;; The compiler likes to be able to directly make value cells.
 (define-vop (make-value-cell)
@@ -312,9 +508,8 @@
   (:info stack-allocate-p)
   (:node-var node)
   (:generator 10
-    (with-fixed-allocation
-        (result value-cell-widetag value-cell-size node stack-allocate-p)
-      (storew value result value-cell-value-slot other-pointer-lowtag))))
+    (fixed-alloc result value-cell-widetag value-cell-size node stack-allocate-p)
+    (storew value result value-cell-value-slot other-pointer-lowtag)))
 
 ;;;; automatic allocators for primitive objects
 
@@ -327,9 +522,12 @@
 (define-vop (make-funcallable-instance-tramp)
   (:args)
   (:results (result :scs (any-reg)))
+  (:vop-var vop)
   (:generator 1
-    ;; TODO: if in immobile space, use LEA to make position-independent
-    (inst mov result (make-fixup 'funcallable-instance-tramp :assembly-routine))))
+    (let ((tramp (make-fixup 'funcallable-instance-tramp :assembly-routine)))
+      (if (sb!c::code-immobile-p vop)
+          (inst lea result (make-ea :qword :base rip-tn :disp tramp))
+          (inst mov result tramp)))))
 
 (define-vop (fixed-alloc)
   (:args)
@@ -356,16 +554,32 @@
                  ;; XXX: should layout fixups use a name, not a layout object?
                  (make-fixup type :layout))))))))
 
+;;; Allocate a non-vector variable-length object.
+;;; Exactly 4 allocators are rendered via this vop:
+;;;  BIGNUM               (%ALLOCATE-BIGNUM)
+;;;  FUNCALLABLE-INSTANCE (%MAKE-FUNCALLABLE-INSTANCE)
+;;;  CLOSURE              (%COPY-CLOSURE)
+;;;  INSTANCE             (%MAKE-INSTANCE)
+;;; WORDS accounts for the mandatory slots *including* the header.
+;;; EXTRA is the variable payload, also measured in words.
 (define-vop (var-alloc)
   (:args (extra :scs (any-reg)))
   (:arg-types positive-fixnum)
   (:info name words type lowtag)
   (:ignore name)
   (:results (result :scs (descriptor-reg) :from (:eval 1)))
-  (:temporary (:sc any-reg :from :eval :to (:eval 1)) bytes)
-  (:temporary (:sc any-reg :from :eval :to :result) header)
+  (:temporary (:sc unsigned-reg :from :eval :to (:eval 1)) bytes)
+  (:temporary (:sc unsigned-reg :from :eval :to :result) header)
   (:node-var node)
   (:generator 50
+   ;; With the exception of bignums, these objects have effectively
+   ;; 32-bit headers because the high half contains other data.
+   (multiple-value-bind (bytes header extra)
+      (if (= type bignum-widetag)
+          (values bytes header extra)
+          (values (reg-in-size bytes  :dword)
+                  (reg-in-size header :dword)
+                  (reg-in-size extra  :dword)))
     (inst lea bytes
           (make-ea :qword :disp (* (1+ words) n-word-bytes) :index extra
                    :scale (ash 1 (- word-shift n-fixnum-tag-bits))))
@@ -374,45 +588,47 @@
     (inst lea header                    ; (w-1 << 8) | type
           (make-ea :qword :base header
                    :disp (+ (ash -2 n-widetag-bits) type)))
-    (inst and bytes (lognot lowtag-mask))
+    (inst and bytes (lognot lowtag-mask)))
     (pseudo-atomic
-     (allocation result bytes node)
-     (inst lea result (make-ea :byte :base result :disp lowtag))
+     (allocation result bytes node nil lowtag)
      (storew header result 0 lowtag))))
 
 #!+immobile-space
-(macrolet ((def (lisp-name c-name arg-scs &body stuff
-                           &aux (argc (length arg-scs)))
-             `(define-vop (,lisp-name)
-                (:args ,@(if (>= argc 1) `((arg1 :scs ,(first arg-scs) :target c-arg1)))
-                       ,@(if (>= argc 2) `((arg2 :scs ,(second arg-scs) :target c-arg2))))
-                ,@(if (>= argc 1)
-                      '((:temporary (:sc unsigned-reg :from (:argument 0)
-                                     :to :eval :offset rdi-offset) c-arg1)))
-                ,@(if (>= argc 2)
-                      '((:temporary (:sc unsigned-reg :from (:argument 1)
-                                     :to :eval :offset rsi-offset) c-arg2)))
-                (:temporary (:sc unsigned-reg :from :eval :to (:result 0)
-                             :offset rax-offset) c-result)
-                (:results (result :scs (descriptor-reg)))
-                (:generator 50
-                 (pseudo-atomic
-                  ,@(if (>= argc 1) '((move c-arg1 arg1)))
-                  ,@(if (>= argc 2) '((move c-arg2 arg2)))
-                  (inst and rsp-tn -16)
-                  (inst mov temp-reg-tn (make-fixup ,c-name :foreign))
-                  (inst call temp-reg-tn)
-                  ,@stuff
-                  (move result c-result))))))
-  ;; These VOPs are each used in one place only, and deliberately not
-  ;; specified as transforming the function after which they are named.
-  (def alloc-immobile-layout "alloc_layout" ; MAKE-LAYOUT
-       ((descriptor-reg)))
-  (def alloc-immobile-symbol "alloc_sym"    ; MAKE-SYMBOL
-       ((descriptor-reg)))
-  (def alloc-immobile-fdefn  "alloc_fdefn"  ; MAKE-FDEFN
-       ((descriptor-reg)))
-  #!+(and immobile-code compact-instance-header)
-  (def alloc-generic-function  "alloc_generic_function"
-       ((descriptor-reg)))
-  )
+(progn
+(define-vop (alloc-immobile-fixedobj)
+  (:info lowtag size word0 word1)
+  (:temporary (:sc unsigned-reg :to :eval :offset rdi-offset) c-arg1)
+  (:temporary (:sc unsigned-reg :to :eval :offset rsi-offset) c-arg2)
+  (:temporary (:sc unsigned-reg :from :eval :to (:result 0) :offset rax-offset)
+              c-result)
+  (:results (result :scs (descriptor-reg)))
+  (:generator 50
+   (inst mov c-arg1 size)
+   (inst mov c-arg2 word0)
+   ;; RSP needn't be restored because the allocators all return immediately
+   ;; which has that effect
+   (inst and rsp-tn -16)
+   (pseudo-atomic
+     (inst call (make-fixup "alloc_fixedobj" :foreign))
+     (inst lea result (make-ea :qword :base c-result :disp lowtag))
+     ;; If code, the next word must be set within the P-A
+     ;; otherwise the GC would compute the wrong object size.
+     (when word1
+       (inst mov (make-ea :qword :base result :disp (- n-word-bytes lowtag)) word1)))))
+(define-vop (alloc-immobile-layout)
+  (:args (slots :scs (descriptor-reg) :target c-arg1))
+  (:temporary (:sc unsigned-reg :from (:argument 0) :to :eval :offset rdi-offset)
+              c-arg1)
+  (:temporary (:sc unsigned-reg :from :eval :to (:result 0) :offset rax-offset)
+              c-result)
+  (:results (result :scs (descriptor-reg)))
+  (:generator 50
+   (move c-arg1 slots)
+   ;; RSP needn't be restored because the allocators all return immediately
+   ;; which has that effect
+   (inst and rsp-tn -16)
+   (pseudo-atomic
+     (inst call (make-fixup "alloc_layout" :foreign)))
+     (move result c-result)))
+) ; end PROGN
+
