@@ -21,42 +21,39 @@
   (:generator 1
    (loadw result object offset lowtag)))
 
+;; This vop is selected by name from vm-ir2tran for converting any
+;; setter or setf'er that is defined by 'objdef'
 (define-vop (set-slot)
   (:args (object :scs (descriptor-reg))
          (value :scs (descriptor-reg any-reg immediate)))
-  (:temporary (:sc descriptor-reg) temp)
   (:info name offset lowtag)
  ;(:ignore name)
   (:results)
   (:generator 1
     (progn name) ; ignore it
-    (generate-set-slot object value temp offset lowtag)))
-
-(defun generate-set-slot (object value temp offset lowtag &optional zeroed)
-    (if (sc-is value immediate)
-        (move-immediate (make-ea :qword
-                                 :base object
-                                 :disp (- (* offset n-word-bytes)
-                                          lowtag))
-                        (encode-value-if-immediate value)
-                        temp zeroed)
-        ;; Else, value not immediate.
-        (storew value object offset lowtag)))
+    (gen-cell-set (make-ea-for-object-slot object offset lowtag) value nil)))
 
 ;; INIT-SLOT has to know about the :COMPACT-INSTANCE-HEADER feature.
 (define-vop (init-slot set-slot)
   (:info name dx-p offset lowtag)
   (:generator 1
     (progn name dx-p)
-    (if (or #!+compact-instance-header
-            (and (eq name '%make-structure-instance) (eql offset :layout)))
+    (cond #!+compact-instance-header
+          ((and (eq name '%make-structure-instance) (eql offset :layout))
         ;; The layout is in the upper half of the header word.
-        (inst mov
+           (inst mov
               (make-ea :dword :base object :disp (- 4 instance-pointer-lowtag))
               (if (sc-is value immediate)
                   (make-fixup (tn-value value) :layout)
-                  (reg-in-size value :dword)))
-        (generate-set-slot object value temp offset lowtag (not dx-p)))))
+                  (reg-in-size value :dword))))
+          ((sc-is value immediate)
+           (move-immediate (make-ea :qword
+                                    :base object
+                                    :disp (- (* offset n-word-bytes) lowtag))
+                           (encode-value-if-immediate value)
+                           temp-reg-tn (not dx-p)))
+          (t
+           (storew value object offset lowtag))))) ; Else, value not immediate.
 
 (define-vop (compare-and-swap-slot)
   (:args (object :scs (descriptor-reg) :to :eval)
@@ -77,13 +74,28 @@
 
 ;;;; symbol hacking VOPs
 
-(define-vop (%set-symbol-global-value cell-set)
-  (:variant symbol-value-slot other-pointer-lowtag))
+(define-vop (%set-symbol-global-value)
+  (:args (object :scs (descriptor-reg immediate))
+         (value :scs (descriptor-reg any-reg immediate)))
+  (:policy :fast-safe)
+  (:generator 4
+    (gen-cell-set (cond ((sc-is object immediate)
+                         (symbol-slot-ea (tn-value object) symbol-value-slot))
+                        (t
+                         (make-ea-for-object-slot object symbol-value-slot
+                                                  other-pointer-lowtag)))
+                  value nil)))
 
-(define-vop (fast-symbol-global-value cell-ref)
-  (:variant symbol-value-slot other-pointer-lowtag)
+(define-vop (fast-symbol-global-value)
+  (:args (object :scs (descriptor-reg immediate)))
+  (:results (value :scs (descriptor-reg any-reg)))
   (:policy :fast)
-  (:translate sym-global-val))
+  (:translate sym-global-val)
+  (:generator 4
+    (cond ((sc-is object immediate)
+           (inst mov value (symbol-slot-ea (tn-value object) symbol-value-slot)))
+          (t
+           (loadw value object symbol-value-slot other-pointer-lowtag)))))
 
 (define-vop (symbol-global-value)
   (:policy :fast-safe)
@@ -98,30 +110,12 @@
       (inst cmp (reg-in-size value :dword) unbound-marker-widetag)
       (inst jmp :e err-lab))))
 
-;; Return T if SYMBOL will be (at no later than load time) wired to a TLS index.
-;; When a symbol gets a tls index, it is permanently at that index, so it is
-;; by some definition "wired", but here the term specifically means that an
-;; index can/should be encoded into the instruction for reading that symbol's
-;; thread-local value, as contrasted with reading its index from itself.
-;; The BIND vop forces wiring; the BOUNDP vop never does, the SET vop doesn't
-;; but could, and {FAST-}SYMBOL-VALUE vops optimize their output appropriately
-;; based on wiring. Except when a symbol maps into a thread slot, the compiler
-;; doesn't care what the actual index is - that's the loader's purview.
-(defun wired-tls-symbol-p (symbol)
-  (not (null (info :variable :wired-tls symbol))))
-
 ;; Return the DISP field to use in an EA relative to thread-base-tn
 (defun load-time-tls-offset (symbol)
   (let ((where (info :variable :wired-tls symbol)))
     (cond ((integerp where) where)
           (t (make-fixup symbol :symbol-tls-index)))))
 
-;; Return T if reference to symbol doesn't need to check for no-tls-value.
-;; True of thread struct slots. More generally we could add a declaration.
-(defun symbol-always-thread-local-p (symbol)
-  (let ((where (info :variable :wired-tls symbol)))
-    (or (eq where :ALWAYS-THREAD-LOCAL) (integerp where))))
-
 (deftransform %compare-and-swap-symbol-value ((symbol old new)
                                               ((constant-arg symbol) t t))
   (if (eq (info :variable :kind (sb!c::lvar-value symbol)) :global)
@@ -145,10 +139,7 @@
                       no-tls-value-marker-widetag)
                 (inst cmov :e cell symbol))) ; now possibly get the symbol
            (access-wired-tls-val (sym) ; SYM is a symbol
-             `(make-ea :qword :disp (load-time-tls-offset ,sym)
-                       :base thread-base-tn))
-           (access-tls-val (index size)
-             `(make-ea ,size :base thread-base-tn :index ,index :scale 1))
+             `(thread-tls-ea (load-time-tls-offset ,sym)))
            (access-value-slot (sym &optional (size :qword)) ; SYM is a TN
              (ecase size
                (:dword `(make-ea-for-object-slot-half
@@ -188,8 +179,7 @@
 
   (define-vop (%cas-symbol-global-value)
     (:translate %cas-symbol-global-value)
-    (:args (symbol :scs (descriptor-reg) :to (:result 0)
-                   :load-if (not (sc-is symbol immediate)))
+    (:args (symbol :scs (descriptor-reg immediate) :to (:result 0))
            (old :scs (descriptor-reg any-reg) :target rax)
            (new :scs (descriptor-reg any-reg)))
     (:temporary (:sc descriptor-reg :offset rax-offset
@@ -200,7 +190,7 @@
       (move rax old)
       (inst cmpxchg
             (if (sc-is symbol immediate)
-                (symbol-slot-addr (tn-value symbol) symbol-value-slot)
+                (symbol-slot-ea (tn-value symbol) symbol-value-slot)
                 (access-value-slot symbol))
             new :lock)
       (move result rax)))
@@ -210,57 +200,97 @@
     ;; TODO: SET could be shorter for any known wired-tls symbol.
     (define-vop (set)
       (:args (symbol :scs (descriptor-reg))
-             (value :scs (descriptor-reg any-reg)
-                    :load-if (or (not (sc-is value immediate))
-                                 (not (typep (encode-value-if-immediate value)
-                                             '(or (signed-byte 32) fixup))))))
+             (value :scs (descriptor-reg any-reg immediate)))
       (:temporary (:sc descriptor-reg) cell)
       (:generator 4
         ;; Compute the address into which to store. CMOV can only move into
         ;; a register, so we can't conditionally move into the TLS and
         ;; conditionally move in the opposite flag sense to the symbol.
         (compute-virtual-symbol)
-        (inst mov (access-value-slot cell)
-              (encode-value-if-immediate value))))
+        (gen-cell-set (access-value-slot cell) value nil)))
+
+    ;; This code is tested by 'codegen.impure.lisp'
+    (defun emit-symeval (value symbol symbol-reg check-boundp vop)
+      (let* ((known-symbol-p (sc-is symbol constant immediate))
+             (known-symbol (and known-symbol-p (tn-value symbol))))
+        ;; In order from best to worst.
+        (cond
+         ((symbol-always-has-tls-value-p known-symbol) ; e.g. *HANDLER-CLUSTERS*
+          (inst mov value (access-wired-tls-val known-symbol)))
+         (t
+          (cond
+            ((symbol-always-has-tls-index-p known-symbol) ; e.g. CL:*PRINT-BASE*
+             ;; Known nonzero TLS index, but possibly no per-thread value.
+             ;; The TLS value and global value can be loaded independently.
+             (inst mov value (access-wired-tls-val known-symbol))
+             (when (sc-is symbol constant)
+               (inst mov symbol-reg symbol))) ; = MOV Rxx, [RIP-N]
+
+            (known-symbol-p ; unknown TLS index, possibly 0
+             (sc-case symbol
+               (immediate
+               ;; load the TLS index from the symbol. TODO: use [RIP-n] mode
+               ;; for immobile code to make it automatically relocatable.
+               (inst mov (reg-in-size value :dword)
+                     ;; slot index 1/2 is the high half of the header word.
+                     (symbol-slot-ea known-symbol 1/2 :dword))
+               ;; read the TLS value using that index
+               (inst mov value (thread-tls-ea value :qword)))
+               (constant
+
+               ;; These reads are inextricably data-dependent
+               (inst mov symbol-reg symbol) ; = MOV REG, [RIP-N]
+               (inst mov (reg-in-size value :dword) (tls-index-of symbol-reg))
+               (inst mov value (thread-tls-ea value :qword)))))
+
+            (t ; SYMBOL-VALUE of a random symbol
+             (inst mov (reg-in-size symbol-reg :dword) (tls-index-of symbol))
+             (inst mov value (thread-tls-ea symbol-reg :qword))
+             (setq symbol-reg symbol)))
+
+          ;; Load the global value if the TLS value didn't exist
+          (inst cmp (reg-in-size value :dword) no-tls-value-marker-widetag)
+          (inst cmov :e value
+                (if (and known-symbol-p (sc-is symbol immediate))
+                    (symbol-slot-ea known-symbol symbol-value-slot) ; MOV Rxx, imm32
+                    (access-value-slot symbol-reg)))))
+
+        (when check-boundp
+          (assemble ()
+            (inst cmp (reg-in-size value :dword) unbound-marker-widetag)
+            (let* ((immediatep (sc-is symbol immediate))
+                   (staticp (and immediatep (static-symbol-p known-symbol)))
+                   (*location-context* (make-restart-location RETRY value)))
+              ;; IMMEDIATE sc symbols are not in a register (they are accessed
+              ;; via absolute address), nor are they present in the code header.
+              ;; So emit a MOV just before the INT opcode for such symbols,
+              ;; out of the normal execution path. Most static symbols are
+              ;; DEFCONSTANTs or DEFGLOBALs, so this case is infrequent.
+              (inst jmp :e (generate-error-code+
+                            (if staticp
+                                (lambda ()
+                                  (load-immediate vop symbol symbol-reg)))
+                            vop 'unbound-symbol-error
+                            (if (and immediatep (not staticp))
+                                symbol
+                                symbol-reg))))
+            RETRY))))
 
   ;; With Symbol-Value, we check that the value isn't the trap object. So
   ;; Symbol-Value of NIL is NIL.
     (define-vop (symbol-value)
       (:translate symeval)
       (:policy :fast-safe)
-      (:args (symbol :scs (descriptor-reg constant) :to (:result 1)))
+      (:args (symbol :scs (descriptor-reg constant immediate) :to (:result 1)))
+      ;; TODO: use no temp if the symbol is known to be thread-local
+      ;; (probably IR1 should go SYMBOL-VALUE -> SYMBOL-TLS-VALUE)
       (:temporary (:sc descriptor-reg) symbol-reg)
       (:results (value :scs (descriptor-reg any-reg)))
       (:vop-var vop)
       (:save-p :compute-only)
       (:variant-vars check-boundp)
       (:variant t)
-      (:generator 9
-        (cond
-          ((not (sc-is symbol constant)) ; SYMBOL-VALUE of a random symbol
-           ;; These reads are inextricably data-dependent
-           (inst mov (reg-in-size symbol-reg :dword) (tls-index-of symbol))
-           (inst mov value (access-tls-val symbol-reg :qword))
-           (setq symbol-reg symbol))
-          ((wired-tls-symbol-p (tn-value symbol)) ; e.g. CL:*PRINT-BASE*
-           (inst mov symbol-reg symbol) ; = MOV Rxx, [RIP-N]
-           (inst mov value (access-wired-tls-val (tn-value symbol))))
-          (t ; commonest case: constant-tn for the symbol, not wired tls
-           ;; Same data-dependencies as the non-constant-tn case.
-           (inst mov symbol-reg symbol) ; = MOV REG, [RIP-N]
-           (inst mov (reg-in-size value :dword) (tls-index-of symbol-reg))
-           (inst mov value (access-tls-val value :qword))))
-        (unless (and (sc-is symbol constant)
-                     (symbol-always-thread-local-p (tn-value symbol)))
-          (inst cmp (reg-in-size value :dword) no-tls-value-marker-widetag)
-          (inst cmov :e value (access-value-slot symbol-reg)))
-        (when check-boundp
-          (assemble ()
-            (inst cmp (reg-in-size value :dword) unbound-marker-widetag)
-            (let ((*location-context* (make-restart-location RETRY value)))
-              (inst jmp :e (generate-error-code vop 'unbound-symbol-error
-                                                symbol-reg)))
-            RETRY))))
+      (:generator 9 (emit-symeval value symbol symbol-reg check-boundp vop)))
 
     (define-vop (fast-symbol-value symbol-value)
     ;; KLUDGE: not really fast, in fact, because we're going to have to
@@ -272,44 +302,6 @@
       (:variant nil)
       (:variant-cost 5))
 
-    ;; SYMBOL-VALUE of a static symbol with a wired TLS index does not
-    ;; load the symbol except in the case of error, and uses no temp either.
-    ;; There's no way to express the lack of encumbrances in the general vop.
-    (define-vop (symeval/static-wired)
-      (:translate symeval)
-      (:policy :fast-safe)
-      ;; The predicates are orthogonal. Symbols can satisfy one, the other,
-      ;; both, or neither. This vop applies only if both.
-      (:arg-types (:constant (and (satisfies static-symbol-p)
-                                  (satisfies wired-tls-symbol-p))))
-      (:info symbol)
-      (:results (value :scs (descriptor-reg any-reg)))
-      (:vop-var vop)
-      (:save-p :compute-only)
-      (:variant-vars check-boundp)
-      (:variant t)
-      (:generator 5
-        (inst mov value (access-wired-tls-val symbol))
-        (unless (symbol-always-thread-local-p symbol)
-          (inst cmp (reg-in-size value :dword) no-tls-value-marker-widetag)
-          (inst cmov :e value (static-symbol-value-ea symbol)))
-        (when check-boundp
-          (let ((err-label (gen-label)))
-            (assemble (*elsewhere*)
-              (emit-label err-label)
-              (inst mov value (+ nil-value (static-symbol-offset symbol)))
-              (emit-error-break vop error-trap
-                                (error-number-or-lose 'unbound-symbol-error)
-                                (list value)))
-            (inst cmp (reg-in-size value :dword) unbound-marker-widetag)
-            (inst jmp :e err-label)))))
-
-    (define-vop (fast-symeval/static-wired symeval/static-wired)
-      (:policy :fast)
-      (:variant nil)
-      (:variant-cost 3))
-
-    ;; Would it be worthwhile to make a static/wired boundp vop?
     (define-vop (boundp)
       (:translate boundp)
       (:policy :fast-safe)
@@ -318,7 +310,7 @@
       (:temporary (:sc dword-reg) temp)
       (:generator 9
         (inst mov temp (tls-index-of object))
-        (inst mov temp (access-tls-val temp :dword))
+        (inst mov temp (thread-tls-ea temp :dword))
         (inst cmp temp no-tls-value-marker-widetag)
         (inst cmov :e temp (access-value-slot object :dword))
         (inst cmp temp unbound-marker-widetag))))
@@ -384,18 +376,14 @@
   (:args (function :scs (descriptor-reg) :target result)
          (fdefn :scs (descriptor-reg)))
   (:temporary (:sc unsigned-reg) raw)
-  (:temporary (:sc unsigned-reg) type)
   (:results (result :scs (descriptor-reg)))
   (:generator 38
-    (load-type type function (- fun-pointer-lowtag))
-    (inst lea raw
-          (make-ea :byte :base function
-                   :disp (- (* simple-fun-code-offset n-word-bytes)
-                            fun-pointer-lowtag)))
-    (inst cmp (reg-in-size type :byte) simple-fun-widetag)
-    (inst jmp :e NORMAL-FUN)
     (inst mov raw (make-fixup 'closure-tramp :assembly-routine))
-    NORMAL-FUN
+    (inst cmp (make-ea :byte :base function :disp (- fun-pointer-lowtag))
+          simple-fun-widetag)
+    (inst cmov :e raw
+          (make-ea :qword :base function
+                   :disp (- (* simple-fun-self-slot n-word-bytes) fun-pointer-lowtag)))
     (storew function fdefn fdefn-fun-slot other-pointer-lowtag)
     (storew raw fdefn fdefn-raw-addr-slot other-pointer-lowtag)
     (move result function)))
@@ -405,12 +393,14 @@
   (:translate fdefn-makunbound)
   (:args (fdefn :scs (descriptor-reg) :target result))
   (:results (result :scs (descriptor-reg)))
+  (:vop-var vop)
   #!+immobile-code (:temporary (:sc unsigned-reg) temp)
   (:generator 38
     #!+immobile-code
-    (progn
-     (inst mov (reg-in-size temp :dword)
-           (make-fixup 'undefined-tramp :assembly-routine))
+    (let ((tramp (make-fixup 'undefined-tramp-fdefn :assembly-routine)))
+     (if (sb!c::code-immobile-p vop)
+         (inst lea temp (make-ea :qword :base rip-tn :disp tramp))
+         (inst mov temp tramp))
      ;; Compute displacement from the call site
      (inst sub (reg-in-size temp :dword) (reg-in-size fdefn :dword))
      (inst sub (reg-in-size temp :dword)
@@ -451,12 +441,12 @@
     (inst test (reg-in-size tls-index :dword) (reg-in-size tls-index :dword))
     (inst jmp :ne TLS-INDEX-VALID)
     (inst mov tls-index symbol)
-    (invoke-asm-routine 'call 'alloc-tls-index vop tmp)
+    (invoke-asm-routine 'call 'alloc-tls-index vop)
     TLS-INDEX-VALID
-    (inst mov tmp (make-ea :qword :base thread-base-tn :index tls-index))
+    (inst mov tmp (thread-tls-ea tls-index))
     (storew tmp bsp (- binding-value-slot binding-size))
     (storew tls-index bsp (- binding-symbol-slot binding-size))
-    (inst mov (make-ea :qword :base thread-base-tn :index tls-index) val)))
+    (inst mov (thread-tls-ea tls-index) val)))
 
 (define-vop (bind) ; bind a known symbol
   (:args (val :scs (any-reg descriptor-reg)))
@@ -465,12 +455,11 @@
   (:generator 10
     (inst mov bsp (* binding-size n-word-bytes))
     (inst xadd
-          (make-ea :qword :base thread-base-tn
-                   :disp (ash thread-binding-stack-pointer-slot word-shift))
+          (thread-tls-ea (ash thread-binding-stack-pointer-slot word-shift))
           bsp)
     (let* ((tls-index (load-time-tls-offset symbol))
-           (tls-cell (make-ea :qword :base thread-base-tn :disp tls-index)))
-      ;; Too bad we can't use "XCHG [r12+disp], val" to write the new value
+           (tls-cell (thread-tls-ea tls-index)))
+      ;; Too bad we can't use "XCHG [thread + disp], val" to write new value
       ;; and read the old value in one step. It will violate the constraints
       ;; prescribed in the internal documentation on special binding.
       (inst mov tmp tls-cell)
@@ -497,23 +486,21 @@
 
 #!+sb-thread
 (define-vop (unbind)
-  (:temporary (:sc unsigned-reg) temp bsp tls-index)
+  (:temporary (:sc unsigned-reg) temp bsp)
   (:temporary (:sc complex-double-reg) zero)
-  (:info n)
+  (:info symbols)
   (:generator 0
     (load-binding-stack-pointer bsp)
     (inst xorpd zero zero)
-    (loop repeat n
+    (loop for symbol in symbols
+          for tls-index = (load-time-tls-offset symbol)
+          for tls-cell = (thread-tls-ea tls-index)
           do
           (inst sub bsp (* binding-size n-word-bytes))
-          ;; Load TLS-INDEX of the SYMBOL from stack
-          (inst mov (reg-in-size tls-index :dword)
-                (make-ea :dword :base bsp :disp (* binding-symbol-slot n-word-bytes)))
 
           ;; Load VALUE from stack, then restore it to the TLS area.
           (loadw temp bsp binding-value-slot)
-          (inst mov (make-ea :qword :base thread-base-tn :index tls-index)
-                temp)
+          (inst mov tls-cell temp)
           ;; Zero out the stack.
           (inst movapd (make-ea :qword :base bsp) zero))
     (store-binding-stack-pointer bsp)))
@@ -558,8 +545,7 @@
     #!-sb-thread
     (storew value symbol symbol-value-slot other-pointer-lowtag)
     #!+sb-thread
-    (inst mov (make-ea :qword :base thread-base-tn :index symbol)
-          value)
+    (inst mov (thread-tls-ea symbol) value)
 
     SKIP
     (inst movapd (make-ea :qword :base bsp) zero)
@@ -584,11 +570,19 @@
   funcallable-instance-info-offset fun-pointer-lowtag
   (descriptor-reg any-reg) * %funcallable-instance-info)
 
-(define-vop (closure-ref slot-ref)
-  (:variant closure-info-offset fun-pointer-lowtag))
+(define-vop (closure-ref)
+  (:args (object :scs (descriptor-reg)))
+  (:results (value :scs (descriptor-reg any-reg)))
+  (:info offset)
+  (:generator 4
+    (loadw value object (+ closure-info-offset offset) fun-pointer-lowtag)))
 
-(define-vop (closure-init slot-set)
-  (:variant closure-info-offset fun-pointer-lowtag))
+(define-vop (closure-init)
+  (:args (object :scs (descriptor-reg))
+         (value :scs (descriptor-reg any-reg)))
+  (:info offset)
+  (:generator 4
+    (storew value object (+ closure-info-offset offset) fun-pointer-lowtag)))
 
 (define-vop (closure-init-from-fp)
   (:args (object :scs (descriptor-reg)))
@@ -610,13 +604,15 @@
   (:policy :fast-safe)
   (:translate %instance-length)
   (:args (struct :scs (descriptor-reg)))
-  (:results (res :scs (unsigned-reg)))
+  ;; Explicitly fixnumizing admits a 32-bit left-shift which might encode
+  ;; in 1 less byte. (Other backends return an UNSIGNED-REG here.)
+  (:results (res :scs (any-reg)))
   (:result-types positive-fixnum)
   (:generator 4
-    (inst mov (reg-in-size res :word)
-              (make-ea :word :base struct
-                       :disp (1+ (- instance-pointer-lowtag))))
-    (inst movzx (reg-in-size res :dword) (reg-in-size res :word))))
+    (let ((res (reg-in-size res :dword)))
+      (inst movzx res (make-ea :word :base struct
+                               :disp (1+ (- instance-pointer-lowtag))))
+      (inst shl res n-fixnum-tag-bits))))
 
 #!+compact-instance-header
 (progn
@@ -651,7 +647,7 @@
   instance-pointer-lowtag (any-reg descriptor-reg) * %instance-ref)
 
 (define-full-setter instance-index-set * instance-slots-offset
-  instance-pointer-lowtag (any-reg descriptor-reg) * %instance-set)
+  instance-pointer-lowtag (any-reg descriptor-reg immediate) * %instance-set)
 
 (define-full-compare-and-swap %instance-cas instance
   instance-slots-offset instance-pointer-lowtag
@@ -780,3 +776,75 @@
     (:generator 4
       (inst xadd (make-ea-for-raw-slot object index) diff :lock)
       (move result diff))))
+
+;;;;
+
+(defknown %cons-cas-pair (cons t t t t) (values t t))
+(defknown %vector-cas-pair (simple-vector index t t t t) (values t t))
+;; %INSTANCE-CAS-PAIR only operates on tagged slots (for now)
+(defknown %instance-cas-pair (instance index t t t t) (values t t))
+
+(defun generate-dblcas (memory-operand old-lo old-hi new-lo new-hi
+                        rax rbx rcx rdx result-lo result-hi)
+  (move rax old-lo)
+  (move rdx old-hi)
+  (move rbx new-lo)
+  (move rcx new-hi)
+  (inst cmpxchg16b memory-operand :lock)
+  ;; RDX:RAX hold the actual old contents of memory.
+  ;; Manually analyze result lifetimes to avoid clobbering.
+  (cond ((and (location= result-lo rdx) (location= result-hi rax))
+         (inst xchg rax rdx)) ; unlikely, but possible
+        ((location= result-lo rdx) ; result-hi is not rax
+         (move result-hi rdx) ; move high part first
+         (move result-lo rax))
+        (t                    ; result-lo is not rdx
+         (move result-lo rax) ; move low part first
+         (move result-hi rdx))))
+
+(macrolet
+    ((define-cmpxchg-vop (name memory-operand more-stuff &optional index-arg)
+       `(define-vop (,name)
+          (:policy :fast-safe)
+          ,@more-stuff
+          (:args (object :scs (descriptor-reg) :to :eval)
+                 ,@index-arg
+                 (expected-old-lo :scs (descriptor-reg any-reg) :target eax)
+                 (expected-old-hi :scs (descriptor-reg any-reg) :target edx)
+                 (new-lo :scs (descriptor-reg any-reg) :target ebx)
+                 (new-hi :scs (descriptor-reg any-reg) :target ecx))
+          (:results (result-lo :scs (descriptor-reg any-reg))
+                    (result-hi :scs (descriptor-reg any-reg)))
+          (:temporary (:sc unsigned-reg :offset eax-offset
+                       :from (:argument 2) :to (:result 0)) eax)
+          (:temporary (:sc unsigned-reg :offset edx-offset
+                       :from (:argument 3) :to (:result 0)) edx)
+          (:temporary (:sc unsigned-reg :offset ebx-offset
+                       :from (:argument 4) :to (:result 0)) ebx)
+          (:temporary (:sc unsigned-reg :offset ecx-offset
+                       :from (:argument 5) :to (:result 0)) ecx)
+          (:generator 7
+           (generate-dblcas ,memory-operand
+                            expected-old-lo expected-old-hi new-lo new-hi
+                            eax ebx ecx edx result-lo result-hi)))))
+  (define-cmpxchg-vop compare-and-exchange-pair
+      (make-ea :dword :base object :disp (- list-pointer-lowtag))
+      ((:translate %cons-cas-pair)))
+  (define-cmpxchg-vop compare-and-exchange-pair-indexed
+      (make-ea :dword :base object :disp offset :index index
+                      :scale (ash n-word-bytes (- n-fixnum-tag-bits)))
+      ((:variant-vars offset))
+      ((index :scs (descriptor-reg any-reg) :to :eval))))
+
+;; The CPU requires 16-byte alignment for the memory operand.
+;; A vector's data portion starts on a 16-byte boundary,
+;; so any even numbered index is OK.
+(define-vop (%vector-cas-pair compare-and-exchange-pair-indexed)
+  (:translate %vector-cas-pair)
+  (:variant (- (* n-word-bytes vector-data-offset) other-pointer-lowtag)))
+
+;; Here you specify an odd numbered slot, otherwise get a bus error.
+;; An instance's first user-visible slot at index 1 is 16-byte-aligned.
+(define-vop (%instance-cas-pair compare-and-exchange-pair-indexed)
+  (:translate %instance-cas-pair)
+  (:variant (- (* n-word-bytes instance-slots-offset) instance-pointer-lowtag)))
