@@ -27,6 +27,7 @@
 #include "os.h"
 #include "arch.h"
 #include "interr.h"
+#include "immobile-space.h"
 #if defined(LISP_FEATURE_OS_PROVIDES_DLOPEN) && !defined(LISP_FEATURE_WIN32)
 # include <dlfcn.h>
 #endif
@@ -131,12 +132,6 @@ os_sem_destroy(os_sem_t *sem)
 
 #endif
 
-#if defined(LISP_FEATURE_OS_PROVIDES_DLOPEN) && !defined(LISP_FEATURE_WIN32)
-void* os_dlopen(char* name, int flags) {
-    return dlopen(name,flags);
-}
-#endif
-
 /* When :SB-DYNAMIC-CORE is enabled, the special category of /static/ foreign
  * symbols disappears. Foreign fixups are resolved to linkage table locations
  * during genesis, and for each of them a record is added to
@@ -158,10 +153,9 @@ os_dlsym_default(char *name)
 }
 #endif
 
+int lisp_linkage_table_n_prelinked;
 void os_link_runtime()
 {
-    extern void write_protect_immobile_space();
-
 #ifdef LISP_FEATURE_SB_DYNAMIC_CORE
     char *link_target = (char*)(intptr_t)LINKAGE_TABLE_SPACE_START;
     void *validated_end = link_target;
@@ -171,12 +165,15 @@ void os_link_runtime()
     void* result;
     int n = 0, m = 0, j;
 
+    if (lisp_linkage_table_n_prelinked)
+        return; // Linkage was already performed by coreparse
+
     struct vector* symbols = VECTOR(SymbolValue(REQUIRED_FOREIGN_SYMBOLS,0));
-    n = fixnum_value(symbols->length);
+    n = lisp_linkage_table_n_prelinked = fixnum_value(symbols->length);
     for (j = 0 ; j < n ; ++j)
     {
         lispobj item = symbols->data[j];
-        datap = lowtag_of(item) == LIST_POINTER_LOWTAG;
+        datap = listp(item);
         symbol_name = datap ? CONS(item)->car : item;
         namechars = (void*)(intptr_t)(VECTOR(symbol_name)->data);
         result = os_dlsym_default(namechars);
@@ -189,10 +186,7 @@ void os_link_runtime()
 #endif
         }
         if (result) {
-            if (datap)
-                arch_write_linkage_table_ref(link_target,result);
-            else
-                arch_write_linkage_table_jmp(link_target,result);
+            arch_write_linkage_table_entry(link_target, result, datap);
         } else {
             m++;
         }
@@ -202,18 +196,6 @@ void os_link_runtime()
     odxprint(runtime_link, "%d total symbols linked, %d undefined",
              n, m);
 #endif /* LISP_FEATURE_SB_DYNAMIC_CORE */
-
-#ifdef LISP_FEATURE_X86_64
-    extern void arch_os_link_runtime();
-    arch_os_link_runtime();
-#endif
-
-#ifdef LISP_FEATURE_IMMOBILE_SPACE
-    /* Delayed until after dynamic space has been mapped, fixups made,
-     * and/or immobile-space linkage entries written,
-     * since it was too soon earlier to handle write faults. */
-    write_protect_immobile_space();
-#endif
 }
 
 #ifndef LISP_FEATURE_WIN32
@@ -238,15 +220,15 @@ gc_managed_addr_p(lispobj ad)
 
     if ((READ_ONLY_SPACE_START <= ad && ad < READ_ONLY_SPACE_END)
         || (STATIC_SPACE_START <= ad && ad < STATIC_SPACE_END)
-#if defined LISP_FEATURE_IMMOBILE_SPACE
-        || (IMMOBILE_SPACE_START <= ad && ad < IMMOBILE_SPACE_END)
-#endif
 #if defined LISP_FEATURE_GENCGC
         || (DYNAMIC_SPACE_START <= ad &&
             ad < (DYNAMIC_SPACE_START + dynamic_space_size))
+        || immobile_space_p(ad)
 #else
-        || (DYNAMIC_0_SPACE_START <= ad && ad < DYNAMIC_0_SPACE_END)
-        || (DYNAMIC_1_SPACE_START <= ad && ad < DYNAMIC_1_SPACE_END)
+        || (DYNAMIC_0_SPACE_START <= ad &&
+            ad < DYNAMIC_0_SPACE_START + dynamic_space_size)
+        || (DYNAMIC_1_SPACE_START <= ad &&
+            ad < DYNAMIC_1_SPACE_START + dynamic_space_size)
 #endif
         )
         return 1;
