@@ -70,8 +70,12 @@
 #include "interr.h"
 #endif
 
-extern char *sbcl_home;
+#ifdef SBCL_PREFIX
 char *sbcl_home = SBCL_PREFIX"/lib/sbcl/";
+#else
+static char libpath[] = "../lib/sbcl";
+char *sbcl_home;
+#endif
 
 #ifdef LISP_FEATURE_HPUX
 extern void *return_from_lisp_stub;
@@ -120,7 +124,7 @@ copied_string(char *string)
     return strcpy(successful_malloc(1+strlen(string)), string);
 }
 
-char *
+static char *
 copied_existing_filename_or_null(char *filename)
 {
     struct stat filename_stat;
@@ -132,7 +136,7 @@ copied_existing_filename_or_null(char *filename)
 }
 
 #ifndef LISP_FEATURE_WIN32
-char *
+static char *
 copied_realpath(const char *pathname)
 {
     char *messy, *tidy;
@@ -164,7 +168,7 @@ copied_realpath(const char *pathname)
 
 /* miscellaneous chattiness */
 
-void
+static void
 print_help()
 {
     puts(
@@ -199,13 +203,13 @@ should be installed along with SBCL, and is also available from the\n\
 website <http://www.sbcl.org/>.\n");
 }
 
-void
+static void
 print_version()
 {
     printf("SBCL %s\n", SBCL_VERSION_STRING);
 }
 
-void
+static void
 print_banner()
 {
     printf(
@@ -230,7 +234,7 @@ and resources this platform demands.\n\
 /* Look for a core file to load, first in the directory named by the
  * SBCL_HOME environment variable, then in a hardcoded default
  * location.  Returns a malloced copy of the core filename. */
-char *
+static char *
 search_for_core ()
 {
     char *env_sbcl_home = getenv("SBCL_HOME");
@@ -259,13 +263,13 @@ search_for_core ()
 /* Try to find the path to an executable from argv[0], this is only
  * used when os_get_runtime_executable_path() returns NULL */
 #ifdef LISP_FEATURE_WIN32
-char *
+static char *
 search_for_executable(const char *argv0)
 {
     return NULL;
 }
 #else /* LISP_FEATURE_WIN32 */
-char *
+static char *
 search_for_executable(const char *argv0)
 {
     char *search, *start, *end, *buf;
@@ -314,7 +318,7 @@ search_for_executable(const char *argv0)
 }
 #endif /* LISP_FEATURE_WIN32 */
 
-size_t
+static size_t
 parse_size_arg(char *arg, char *arg_name)
 {
   char *tail, *power_name;
@@ -360,14 +364,12 @@ parse_size_arg(char *arg, char *arg_name)
 char **posix_argv;
 char *core_string;
 
-struct runtime_options *runtime_options;
-
 char *saved_runtime_path = NULL;
 #if defined(LISP_FEATURE_WIN32) && defined(LISP_FEATURE_SB_THREAD)
 void pthreads_win32_init();
 #endif
 
-void print_locale_variable(const char *name)
+static void print_locale_variable(const char *name)
 {
   char *value = getenv(name);
 
@@ -376,7 +378,7 @@ void print_locale_variable(const char *name)
   }
 }
 
-void setup_locale()
+static void setup_locale()
 {
   if(setlocale(LC_ALL, "") == NULL) {
 #ifndef LISP_FEATURE_WIN32
@@ -410,7 +412,7 @@ void setup_locale()
 #endif
   }
 }
-void print_environment(int argc, char *argv[])
+static void print_environment(int argc, char *argv[])
 {
     int n = 0;
     printf("; Commandline arguments:\n");
@@ -426,6 +428,7 @@ void print_environment(int argc, char *argv[])
     }
 }
 
+extern void write_protect_immobile_space();
 struct lisp_startup_options lisp_startup_options;
 int
 main(int argc, char *argv[], char *envp[])
@@ -444,10 +447,18 @@ main(int argc, char *argv[], char *envp[])
 
     /* other command line options */
     boolean end_runtime_options = 0;
-    boolean disable_lossage_handler_p = 0;
+    boolean disable_lossage_handler_p
+#if defined(LISP_FEATURE_SB_LDB)
+        = 0;
+#else
+        = 1;
+#endif
+
     boolean debug_environment_p = 0;
 
     lispobj initial_function;
+    int merge_core_pages = -1;
+    struct memsize_options memsize_options = {0, 0, 0};
 
 #if defined(LISP_FEATURE_WIN32) && defined(LISP_FEATURE_SB_THREAD)
     os_preinit();
@@ -457,8 +468,6 @@ main(int argc, char *argv[], char *envp[])
     interrupt_init();
     block_blockable_signals(0);
 
-    runtime_options = NULL;
-
     /* Save the argv[0] derived runtime path in case
      * os_get_runtime_executable_path(1) isn't able to get an
      * externally-usable path later on. */
@@ -470,7 +479,8 @@ main(int argc, char *argv[], char *envp[])
     runtime_path = os_get_runtime_executable_path(0);
     if (runtime_path || saved_runtime_path) {
         os_vm_offset_t offset = search_for_embedded_core(
-            runtime_path ? runtime_path : saved_runtime_path);
+            runtime_path ? runtime_path : saved_runtime_path,
+            &memsize_options);
         if (offset != -1) {
             embedded_core_offset = offset;
             core = (runtime_path ? runtime_path :
@@ -484,15 +494,13 @@ main(int argc, char *argv[], char *envp[])
 
     /* Parse our part of the command line (aka "runtime options"),
      * stripping out those options that we handle. */
-    if (runtime_options != NULL) {
-        dynamic_space_size = runtime_options->dynamic_space_size;
-        thread_control_stack_size = runtime_options->thread_control_stack_size;
+    if (memsize_options.present_in_core) {
+        dynamic_space_size = memsize_options.dynamic_space_size;
+        thread_control_stack_size = memsize_options.thread_control_stack_size;
         sbcl_argv = argv;
     } else {
         int argi = 1;
 
-        runtime_options = successful_malloc(sizeof(struct runtime_options));
-
         while (argi < argc) {
             char *arg = argv[argi];
             if (0 == strcmp(arg, "--script")) {
@@ -575,9 +583,6 @@ main(int argc, char *argv[], char *envp[])
             } else if (0 == strcmp(arg, "--no-merge-core-pages")) {
                 ++argi;
                 merge_core_pages = 0;
-            } else if (0 == strcmp(arg, "--default-merge-core-pages")) {
-                ++argi;
-                merge_core_pages = -1;
             } else {
                 /* This option was unrecognized as a runtime option,
                  * so it must be a toplevel option or a user option,
@@ -621,10 +626,6 @@ main(int argc, char *argv[], char *envp[])
 #endif
     thread_control_stack_size &= ~(sword_t)(CONTROL_STACK_ALIGNMENT_BYTES-1);
 
-    /* Preserve the runtime options for possible future core saving */
-    runtime_options->dynamic_space_size = dynamic_space_size;
-    runtime_options->thread_control_stack_size = thread_control_stack_size;
-
     /* KLUDGE: os_vm_page_size is set by os_init(), and on some
      * systems (e.g. Alpha) arch_init() needs need os_vm_page_size, so
      * it must follow os_init(). -- WHN 2000-01-26 */
@@ -641,12 +642,43 @@ main(int argc, char *argv[], char *envp[])
 
     setup_locale();
 
+    #ifndef SBCL_PREFIX
+    /* If built without SBCL_PREFIX defined, then set 'sbcl_home' to
+     * "<here>/../lib/sbcl/" based on how this executable was invoked. */
+    {
+        char *exename = argv[0]; // Use as-it, not truenameified
+        char *slash = strrchr(exename, '/');
+        if (!slash) {
+            sbcl_home = libpath;
+        } else {
+            int prefixlen = slash - exename + 1; // keep the slash in the prefix
+            char *tail = exename + prefixlen - 4;
+            char *suffix = libpath;
+            sbcl_home = successful_malloc(prefixlen + sizeof libpath); // sizeof incl. nul
+            // Translate "{path/}bin/sbcl" => "{path/}lib/sbcl", otherwise
+            // "{path}/sbcl" => "{path}/../lib/sbcl" so that running "./sbcl" works
+            // if sitting in "bin".
+            if (prefixlen >= 4 && !strncmp(tail, "bin/", 4)
+                // chop "bin" only if a complete word: '/' or nothing to its left.
+                && (tail-1 < exename || tail[-1] == '/')) {
+                prefixlen -= 4; // remove "bin/"
+                suffix += 3; // don't append "../"
+            }
+            memcpy(sbcl_home, exename, prefixlen);
+            strcpy(sbcl_home+prefixlen, suffix);
+        }
+    }
+    #endif
+
     /* If no core file was specified, look for one. */
     if (!core) {
         core = search_for_core();
     }
 
-    if (!lisp_startup_options.noinform && embedded_core_offset == 0) {
+    if (embedded_core_offset)
+        lisp_startup_options.noinform = 1;
+
+    if (!lisp_startup_options.noinform) {
         print_banner();
         fflush(stdout);
     }
@@ -659,11 +691,23 @@ main(int argc, char *argv[], char *envp[])
          * before we reach this block, so that there is no observable
          * difference between "embedded" and "bare" images given to
          * --core. */
-        os_vm_offset_t offset = search_for_embedded_core(core);
+        os_vm_offset_t offset = search_for_embedded_core(core, 0);
         if (offset != -1)
             embedded_core_offset = offset;
     }
 
+    globals_init();
+
+    /* Doing this immediately after the core has been located
+     * and before any random malloc() calls occur improves the chance
+     * of mapping dynamic space at our preferred address (if movable).
+     * If not movable, it was already mapped in allocate_spaces(). */
+    initial_function = load_core_file(core, embedded_core_offset,
+                                      merge_core_pages);
+    if (initial_function == NIL) {
+        lose("couldn't find initial function\n");
+    }
+
 #if defined(SVR4) || defined(__linux__) || defined(__NetBSD__)
     tzset();
 #endif
@@ -674,14 +718,12 @@ main(int argc, char *argv[], char *envp[])
     if (!disable_lossage_handler_p)
         enable_lossage_handler();
 
-    globals_init();
-
-    initial_function = load_core_file(core, embedded_core_offset);
-    if (initial_function == NIL) {
-        lose("couldn't find initial function\n");
-    }
-#ifdef LISP_FEATURE_SB_DYNAMIC_CORE
     os_link_runtime();
+#ifdef LISP_FEATURE_IMMOBILE_SPACE
+    /* Delayed until after dynamic space has been mapped, fixups made,
+     * and/or immobile-space linkage entries written,
+     * since it was too soon earlier to handle write faults. */
+    write_protect_immobile_space();
 #endif
 #ifdef LISP_FEATURE_HPUX
     /* -1 = CLOSURE_FUN_OFFSET, 23 = SIMPLE_FUN_CODE_OFFSET, we are
@@ -689,11 +731,6 @@ main(int argc, char *argv[], char *envp[])
     return_from_lisp_stub = (void *) ((char *)*((unsigned long *)
                  ((char *)initial_function + -1)) + 23);
 #endif
-#ifdef LISP_FEATURE_X86_64
-    tune_asm_routines_for_microarch();
-#endif
-
-    gc_initialize_pointers();
 
     arch_install_interrupt_handlers();
 #ifndef LISP_FEATURE_WIN32
