@@ -40,10 +40,10 @@
 #include "dynbind.h"
 #include "genesis/cons.h"
 #include "genesis/fdefn.h"
+#include "genesis/vector.h"
 #include "interr.h"             /* for lose() */
 #include "alloc.h"
 #include "gc-internal.h"
-#include "cpputil.h"
 #include "pseudo-atomic.h"
 #include "interrupt.h"
 #include "lispregs.h"
@@ -371,9 +371,9 @@ init_new_thread(struct thread *th, init_thread_data *scribble, int guardp)
      * non-safepoint versions of this code.  Can we unify this more?
      */
 #ifdef LISP_FEATURE_SB_SAFEPOINT
-    gc_state_lock();
-    gc_state_wait(GC_NONE);
-    gc_state_unlock();
+    WITH_GC_STATE_LOCK {
+        gc_state_wait(GC_NONE);
+    }
     push_gcing_safety(&scribble->safety);
 #endif
 }
@@ -388,9 +388,9 @@ undo_init_new_thread(struct thread *th, init_thread_data *scribble)
      */
 #ifdef LISP_FEATURE_SB_SAFEPOINT
     block_blockable_signals(0);
-    gc_alloc_update_page_tables(BOXED_PAGE_FLAG, &th->alloc_region);
+    ensure_region_closed(&th->alloc_region, BOXED_PAGE_FLAG);
 #if defined(LISP_FEATURE_SB_SAFEPOINT_STRICTLY) && !defined(LISP_FEATURE_WIN32)
-    gc_alloc_update_page_tables(BOXED_PAGE_FLAG, &th->sprof_alloc_region);
+    ensure_region_closed(&th->sprof_alloc_region, BOXED_PAGE_FLAG);
 #endif
     pop_gcing_safety(&scribble->safety);
     lock_ret = pthread_mutex_lock(&all_threads_lock);
@@ -408,9 +408,9 @@ undo_init_new_thread(struct thread *th, init_thread_data *scribble)
     lock_ret = pthread_mutex_lock(&all_threads_lock);
     gc_assert(lock_ret == 0);
 
-    gc_alloc_update_page_tables(BOXED_PAGE_FLAG, &th->alloc_region);
+    ensure_region_closed(&th->alloc_region, BOXED_PAGE_FLAG);
 #if defined(LISP_FEATURE_SB_SAFEPOINT_STRICTLY) && !defined(LISP_FEATURE_WIN32)
-    gc_alloc_update_page_tables(BOXED_PAGE_FLAG, &th->sprof_alloc_region);
+    ensure_region_closed(&th->sprof_alloc_region, BOXED_PAGE_FLAG);
 #endif
     unlink_thread(th);
     pthread_mutex_unlock(&all_threads_lock);
@@ -455,13 +455,13 @@ undo_init_new_thread(struct thread *th, init_thread_data *scribble)
  * lisp function after doing arch_os_thread_init and whatever other
  * bookkeeping needs to be done
  */
-int
-new_thread_trampoline(struct thread *th)
+void* new_thread_trampoline(void* arg)
 {
+    struct thread *th = (struct thread *)arg;
     int result;
     init_thread_data scribble;
 
-    FSHOW((stderr,"/creating thread %lu\n", thread_self()));
+    FSHOW((stderr,"/creating thread %p\n", thread_self()));
     check_deferrables_blocked_or_lose(0);
 #ifndef LISP_FEATURE_SB_SAFEPOINT
     check_gc_signals_unblocked_or_lose(0);
@@ -475,8 +475,8 @@ new_thread_trampoline(struct thread *th)
 
     schedule_thread_post_mortem(th);
 
-    FSHOW((stderr,"/exiting thread %lu\n", thread_self()));
-    return result;
+    FSHOW((stderr,"/exiting thread %p\n", thread_self()));
+    return (void*)(uintptr_t)result;
 }
 
 static struct thread *create_thread_struct(lispobj);
@@ -488,9 +488,16 @@ attach_os_thread(init_thread_data *scribble)
     os_thread_t os = pthread_self();
     odxprint(misc, "attach_os_thread: attaching to %p", os);
 
-    struct thread *th = create_thread_struct(NIL);
+    struct thread *th = create_thread_struct(NO_TLS_VALUE_MARKER_WIDETAG);
     block_deferrable_signals(&scribble->oldset);
-    th->no_tls_value_marker = NO_TLS_VALUE_MARKER_WIDETAG;
+
+#ifndef LISP_FEATURE_SB_SAFEPOINT
+    /* initial-thread-function-trampoline doesn't like when the GC signal is blocked */
+    /* FIXME: could be done using a single call to pthread_sigmask
+       together with locking the deferrable signals above. */
+    unblock_gc_signals(0, 0);
+#endif
+
     /* We don't actually want a pthread_attr here, but rather than add
      * `if's to the post-mostem, let's just keep that code happy by
      * keeping it initialized: */
@@ -512,8 +519,8 @@ attach_os_thread(init_thread_data *scribble)
   stack_size = stack.ss_size;
   stack_addr = (void*)((size_t)stack.ss_sp - stack_size);
 #elif defined(LISP_FEATURE_DARWIN)
-    stack_addr = pthread_get_stackaddr_np(os);
     stack_size = pthread_get_stacksize_np(os);
+    stack_addr = pthread_get_stackaddr_np(os) - stack_size;
 #else
     pthread_attr_t attr;
 #ifdef LISP_FEATURE_FREEBSD
@@ -532,16 +539,6 @@ attach_os_thread(init_thread_data *scribble)
 
     init_new_thread(th, scribble, 0);
 
-    /* We will be calling into Lisp soon, and the functions being called
-     * recklessly ignore the comment in target-thread which says that we
-     * must be careful to not cause GC while initializing a new thread.
-     * Since we first need to create a fresh thread object, it's really
-     * tempting to just perform such unsafe allocation though.  So let's
-     * at least try to suppress GC before consing, and hope that it
-     * works: */
-    // Just stomp on the value already set by create_thread_struct()
-    write_TLS(GC_INHIBIT, T, th);
-
     uword_t stacksize
         = (uword_t) th->control_stack_end - (uword_t) th->control_stack_start;
     odxprint(misc, "attach_os_thread: attached %p as %p (0x%lx bytes stack)",
@@ -562,6 +559,12 @@ detach_os_thread(init_thread_data *scribble)
     free_thread_struct(th);
 }
 
+#if defined(LISP_FEATURE_X86_64) && !defined(LISP_FEATURE_WIN32)
+extern void funcall_alien_callback(lispobj arg1, lispobj arg2, lispobj arg0,
+                                   struct thread* thread)
+  __attribute__((sysv_abi));
+#endif
+
 void
 callback_wrapper_trampoline(
 #if !(defined(LISP_FEATURE_X86) || defined(LISP_FEATURE_X86_64))
@@ -582,7 +585,12 @@ callback_wrapper_trampoline(
     if (!th) {                  /* callback invoked in non-lisp thread */
         init_thread_data scribble;
         attach_os_thread(&scribble);
-        funcall3(StaticSymbolFunction(ENTER_FOREIGN_CALLBACK), arg0,arg1,arg2);
+#ifdef LISP_FEATURE_SB_SAFEPOINT
+        WITH_GC_AT_SAFEPOINTS_ONLY()
+#endif
+        {
+            funcall3(StaticSymbolFunction(ENTER_FOREIGN_CALLBACK), arg0,arg1,arg2);
+        }
         detach_os_thread(&scribble);
         return;
     }
@@ -596,8 +604,11 @@ callback_wrapper_trampoline(
     WITH_GC_AT_SAFEPOINTS_ONLY()
 #endif
     {
-       funcall3(StaticSymbolFunction(ENTER_ALIEN_CALLBACK), arg0, arg1, arg2);
-
+#if defined(LISP_FEATURE_X86_64) && !defined(LISP_FEATURE_WIN32)
+        funcall_alien_callback(arg1, arg2, arg0, th);
+#else
+        funcall3(StaticSymbolFunction(ENTER_ALIEN_CALLBACK), arg0,arg1,arg2);
+#endif
     }
 }
 #endif /* LISP_FEATURE_SB_THREAD */
@@ -612,12 +623,6 @@ free_thread_struct(struct thread *th)
 #endif
 }
 
-#ifdef LISP_FEATURE_SB_THREAD
-/* FIXME: should be MAX_INTERRUPTS -1 ? */
-const unsigned int tls_index_start =
-  MAX_INTERRUPTS + sizeof(struct thread)/sizeof(lispobj);
-#endif
-
 /* this is called from any other thread to create the new one, and
  * initialize all parts of it that can be initialized from another
  * thread
@@ -646,9 +651,7 @@ create_thread_struct(lispobj initial_function) {
         return NULL;
     /* Aligning up is safe as THREAD_STRUCT_SIZE has
      * THREAD_ALIGNMENT_BYTES padding. */
-    aligned_spaces = (void *)((((uword_t)(char *)spaces)
-                               + THREAD_ALIGNMENT_BYTES-1)
-                              &~(uword_t)(THREAD_ALIGNMENT_BYTES-1));
+    aligned_spaces = PTR_ALIGN_UP(spaces, THREAD_ALIGNMENT_BYTES);
     char* csp_page=
         (aligned_spaces+
          thread_control_stack_size+
@@ -674,15 +677,18 @@ create_thread_struct(lispobj initial_function) {
     set_binding_stack_pointer(th,th->binding_stack_start);
     th->this=th;
     th->os_thread=0;
+#if defined(LAYOUT_OF_FUNCTION) && defined(LISP_FEATURE_SB_THREAD)
+    th->function_layout = LAYOUT_OF_FUNCTION << 32;
+#endif
+    // Once allocated, the allocation profiling buffer sticks around.
+    // If present and enabled, assign into the new thread.
+    th->profile_data = (uword_t*)(alloc_profiling ? alloc_profile_buffer : 0);
 
 #ifdef LISP_FEATURE_SB_SAFEPOINT
 # ifdef LISP_FEATURE_WIN32
     th->carried_base_pointer = 0;
 # endif
-# ifdef LISP_FEATURE_C_STACK_IS_CONTROL_STACK
-    th->pc_around_foreign_call = 0;
-# endif
-    th->csp_around_foreign_call = csp_page;
+    th->csp_around_foreign_call = (lispobj *)th - 1;
 #endif
 
     struct nonpointer_thread_data *nonpointer_data
@@ -720,9 +726,9 @@ create_thread_struct(lispobj initial_function) {
 #endif
 
 #ifdef LISP_FEATURE_GENCGC
-    gc_set_region_empty(&th->alloc_region);
+    gc_init_region(&th->alloc_region);
 # if defined(LISP_FEATURE_SB_SAFEPOINT_STRICTLY) && !defined(LISP_FEATURE_WIN32)
-    gc_set_region_empty(&th->sprof_alloc_region);
+    gc_init_region(&th->sprof_alloc_region);
 # endif
 #endif
 #ifdef LISP_FEATURE_SB_THREAD
@@ -756,12 +762,38 @@ create_thread_struct(lispobj initial_function) {
 
     th->interrupt_data->pending_handler = 0;
     th->interrupt_data->gc_blocked_deferrables = 0;
-#ifdef GENCGC_IS_PRECISE
+#if GENCGC_IS_PRECISE
     th->interrupt_data->allocation_trap_context = 0;
 #endif
 
+#ifdef LISP_FEATURE_SB_THREAD
+// This macro is the same as "write_TLS(sym,val,th)" but can't be spelled thus.
+// 'sym' would get substituted prior to token pasting, so you end up with a bad
+// token "(*)_tlsindex" because all symbols are #defined to "(*)" so that #ifdef
+// remains meaningful to the preprocessor, while use of 'sym' itself yields
+// a deliberate syntax error if you try to compile an expression involving it.
+#  define INITIALIZE_TLS(sym,val) write_TLS_index(sym##_tlsindex, val, th, _ignored_)
+#else
+#  define INITIALIZE_TLS(sym,val) SYMBOL(sym)->value = val
+#endif
 #include "genesis/thread-init.inc"
 #ifdef LISP_FEATURE_SB_THREAD
+    /* Each initial binding is a cons whose car is a symbol evaluated as if
+     * by SYMBOL-GLOBAL-VALUE (unsafely), and whose cdr is the target symbol.
+     * In particular, we will obligingly assign the unbound-marker.
+     * An atom implies NIL for the value. */
+    struct vector* tls_init = VECTOR(SYMBOL(THREAD_INITIAL_BINDINGS)->value);
+    for (i = 0; i < tls_init->length; i += make_fixnum(1)) {
+        lispobj binding = tls_init->data[fixnum_value(i)];
+        lispobj value = NIL;
+        if (listp(binding)) {
+            lispobj val_form = CONS(binding)->car;
+            value = fixnump(val_form) ? val_form : SYMBOL(val_form)->value;
+            binding = CONS(binding)->cdr;
+        }
+        struct symbol* sym = SYMBOL(binding);
+        write_TLS_index(tls_index_of(sym), value, th, sym);
+    }
     /* If a symbol assigned above had a TLS index of 0, then it'll
      * mess up th->no_tls_value_marker. Fail now if that happened. */
     gc_assert(th->no_tls_value_marker == NO_TLS_VALUE_MARKER_WIDETAG);
@@ -780,7 +812,7 @@ create_thread_struct(lispobj initial_function) {
 }
 
 void create_initial_thread(lispobj initial_function) {
-    struct thread *th=create_thread_struct(initial_function);
+    struct thread *th = create_thread_struct(initial_function);
 #ifdef LISP_FEATURE_SB_THREAD
     pthread_key_create(&lisp_thread, 0);
 #endif
@@ -830,8 +862,7 @@ boolean create_os_thread(struct thread *th,os_thread_t *kid_tid)
                               ALIEN_STACK_SIZE)) ||
 # endif
 #endif
-       (retcode = pthread_create
-        (kid_tid,th->os_attr,(void *(*)(void *))new_thread_trampoline,th))) {
+       (retcode = pthread_create(kid_tid, th->os_attr, new_thread_trampoline, th))) {
         FSHOW_SIGNAL((stderr, "init = %d\n", initcode));
         FSHOW_SIGNAL((stderr, "pthread_create returned %d, errno %d\n",
                       retcode, errno));
